"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2024 Olivier Friard

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  MA 02110-1301, USA.
"""

from decimal import Decimal as dec
import tablib
import logging
import os
import sys
import datetime as dt
import math
import pathlib
from io import StringIO
import pandas as pd
from typing import Tuple

try:
    import pyreadr

    flag_pyreadr_loaded = True
except ModuleNotFoundError:
    flag_pyreadr_loaded = False

from . import dialog
from . import config as cfg
from . import utilities as util
from . import project_functions
from . import observation_operations
from . import db_functions
from . import event_operations


def export_events_jwatcher(
    parameters: dict, obsId: str, observation: list, ethogram: dict, file_name: str, output_format: str
) -> Tuple[bool, str]:
    """
    export events jwatcher .dat format

    Args:
        parameters (dict): subjects, behaviors
        obsId (str): observation id
        observation (dict): observation
        ethogram (dict): ethogram of project
        file_name (str): file name for exporting events
        output_format (str): Not used for compatibility with export_events function

    Returns:
        bool: result: True if OK else False
        str: error message
    """
    for subject in parameters[cfg.SELECTED_SUBJECTS]:
        # select events for current subject
        events = []
        for event in observation[cfg.EVENTS]:
            if event[cfg.EVENT_SUBJECT_FIELD_IDX] == subject or (
                subject == cfg.NO_FOCAL_SUBJECT and event[cfg.EVENT_SUBJECT_FIELD_IDX] == ""
            ):
                events.append(event)

        if not events:
            continue

        total_length = 0  # in seconds
        if observation[cfg.EVENTS]:
            total_length = observation[cfg.EVENTS][-1][0] - observation[cfg.EVENTS][0][0]  # last event time - first event time

        file_name_subject = str(pathlib.Path(file_name).parent / pathlib.Path(file_name).stem) + "_" + subject + ".dat"

        rows = ["FirstLineOfData"]  # to be completed
        rows.append("#-----------------------------------------------------------")
        rows.append(f"# Name: {pathlib.Path(file_name_subject).name}")
        rows.append("# Format: Focal Data File 1.0")
        rows.append(f"# Updated: {dt.datetime.now().isoformat()}")
        rows.append("#-----------------------------------------------------------")
        rows.append("")
        rows.append(f"FocalMasterFile={pathlib.Path(file_name_subject).with_suffix('.fmf')}")
        rows.append("")

        rows.append(f"# Observation started: {observation['date']}")
        try:
            start_time = dt.datetime.strptime(observation["date"], "%Y-%m-%dT%H:%M:%S")
        except ValueError:
            start_time = dt.datetime(1970, 1, 1, 0, 0)
        start_time_epoch = int((start_time - dt.datetime(1970, 1, 1, 0, 0)).total_seconds() * 1000)
        rows.append(f"StartTime={start_time_epoch}")

        stop_time = (start_time + dt.timedelta(seconds=float(total_length))).isoformat()
        stop_time_epoch = int(start_time_epoch + float(total_length) * 1000)

        rows.append(f"# Observation stopped: {stop_time}")
        rows.append(f"StopTime={stop_time_epoch}")

        rows.extend([""] * 3)
        rows.append("#BEGIN DATA")
        rows[0] = f"FirstLineOfData={len(rows) + 1}"

        all_observed_behaviors = []
        mem_number_of_state_events = {}
        for event in events:
            behav_code = event[cfg.EVENT_BEHAVIOR_FIELD_IDX]

            try:
                behavior_key = [ethogram[k][cfg.BEHAVIOR_KEY] for k in ethogram if ethogram[k][cfg.BEHAVIOR_CODE] == behav_code][0]
            except Exception:
                # coded behavior not defined in ethogram
                continue
            if [ethogram[k][cfg.TYPE] for k in ethogram if ethogram[k][cfg.BEHAVIOR_CODE] == behav_code] == [cfg.STATE_EVENT]:
                if behav_code in mem_number_of_state_events:
                    mem_number_of_state_events[behav_code] += 1
                else:
                    mem_number_of_state_events[behav_code] = 1
                # skip the STOP event in case of STATE
                if mem_number_of_state_events[behav_code] % 2 == 0:
                    continue

            rows.append(f"{int(event[cfg.EVENT_TIME_FIELD_IDX] * 1000)}, {behavior_key}")
            if (event[cfg.EVENT_BEHAVIOR_FIELD_IDX], behavior_key) not in all_observed_behaviors:
                all_observed_behaviors.append((event[cfg.EVENT_BEHAVIOR_FIELD_IDX], behavior_key))

        rows.append(f"{int(events[-1][0] * 1000)}, EOF\n")

        try:
            with open(file_name_subject, "w") as f_out:
                f_out.write("\n".join(rows))
        except Exception:
            return False, f"File DAT not created for subject {subject}: {sys.exc_info()[1]}"

        # create fmf file
        fmf_file_path = pathlib.Path(file_name_subject).with_suffix(".fmf")
        fmf_creation_answer = ""
        if fmf_file_path.exists():
            fmf_creation_answer = dialog.MessageDialog(
                cfg.programName,
                (f"The {fmf_file_path} file already exists.<br>" "What do you want to do?"),
                [cfg.OVERWRITE, "Skip file creation", cfg.CANCEL],
            )

            if fmf_creation_answer == cfg.CANCEL:
                return True, ""

        rows = []
        rows.append("#-----------------------------------------------------------")
        rows.append(f"# Name: {pathlib.Path(file_name_subject).with_suffix('.fmf').name}")
        rows.append("# Format: Focal Master File 1.0")
        rows.append(f"# Updated: {dt.datetime.now().isoformat()}")
        rows.append("#-----------------------------------------------------------")
        for behav, key in all_observed_behaviors:
            rows.append(f"Behaviour.name.{key}={behav}")
            behav_description = [ethogram[k][cfg.DESCRIPTION] for k in ethogram if ethogram[k][cfg.BEHAVIOR_CODE] == behav][0]
            rows.append(f"Behaviour.description.{key}={behav_description}")

        rows.append(f"DurationMilliseconds={int(float(total_length) * 1000)}")
        rows.append("CountUp=false")
        rows.append("Question.1=")
        rows.append("Question.2=")
        rows.append("Question.3=")
        rows.append("Question.4=")
        rows.append("Question.5=")
        rows.append("Question.6=")
        rows.append("Notes=")
        rows.append("Supplementary=\n")

        if fmf_creation_answer == cfg.OVERWRITE or fmf_creation_answer == "":
            try:
                with open(fmf_file_path, "w") as f_out:
                    f_out.write("\n".join(rows))
            except Exception:
                return False, f"File FMF not created: {sys.exc_info()[1]}"

        # create FAF file
        faf_file_path = pathlib.Path(file_name_subject).with_suffix(".faf")
        faf_creation_answer = ""
        if faf_file_path.exists():
            faf_creation_answer = dialog.MessageDialog(
                cfg.programName,
                (f"The {faf_file_path} file already exists.<br>" "What do you want to do?"),
                [cfg.OVERWRITE, "Skip file creation", cfg.CANCEL],
            )
            if faf_creation_answer == cfg.CANCEL:
                return True, ""

        rows = []
        rows.append("#-----------------------------------------------------------")
        rows.append("# Name: {}".format(pathlib.Path(file_name_subject).with_suffix(".faf").name))
        rows.append("# Format: Focal Analysis Master File 1.0")
        rows.append("# Updated: {}".format(dt.datetime.now().isoformat()))
        rows.append("#-----------------------------------------------------------")
        rows.append("FocalMasterFile={}".format(str(pathlib.Path(file_name_subject).with_suffix(".fmf"))))
        rows.append("")
        rows.append("TimeBinDuration=0.0")
        rows.append("EndWithLastCompleteBin=true")
        rows.append("")
        rows.append("ScoreFromBeginning=true")
        rows.append("ScoreFromBehavior=false")
        rows.append("ScoreFromFirstBehavior=false")
        rows.append("ScoreFromOffset=false")
        rows.append("")
        rows.append("Offset=0.0")
        rows.append("BehaviorToScoreFrom=")
        rows.append("")
        rows.append("OutOfSightCode=")
        rows.append("")
        rows.append("Report.StateNaturalInterval.Occurrence=false")
        rows.append("Report.StateNaturalInterval.TotalTime=false")
        rows.append("Report.StateNaturalInterval.Average=false")
        rows.append("Report.StateNaturalInterval.StandardDeviation=false")
        rows.append("Report.StateNaturalInterval.ProportionOfTime=false")
        rows.append("Report.StateNaturalInterval.ProportionOfTimeInSight=false")
        rows.append("Report.StateNaturalInterval.ConditionalProportionOfTime=false")
        rows.append("")
        rows.append("Report.StateNaturalDuration.Occurrence=false")
        rows.append("Report.StateNaturalDuration.TotalTime=false")
        rows.append("Report.StateNaturalDuration.Average=false")
        rows.append("Report.StateNaturalDuration.StandardDeviation=false")
        rows.append("Report.StateNaturalDuration.ProportionOfTime=false")
        rows.append("Report.StateNaturalDuration.ProportionOfTimeInSight=false")
        rows.append("Report.StateNaturalDuration.ConditionalProportionOfTime=false")
        rows.append("")
        rows.append("Report.StateAllInterval.Occurrence=false")
        rows.append("Report.StateAllInterval.TotalTime=false")
        rows.append("Report.StateAllInterval.Average=false")
        rows.append("Report.StateAllInterval.StandardDeviation=false")
        rows.append("Report.StateAllInterval.ProportionOfTime=false")
        rows.append("Report.StateAllInterval.ProportionOfTimeInSight=false")
        rows.append("Report.StateAllInterval.ConditionalProportionOfTime=false")
        rows.append("")
        rows.append("Report.StateAllDuration.Occurrence=true")
        rows.append("Report.StateAllDuration.TotalTime=true")
        rows.append("Report.StateAllDuration.Average=true")
        rows.append("Report.StateAllDuration.StandardDeviation=false")
        rows.append("Report.StateAllDuration.ProportionOfTime=false")
        rows.append("Report.StateAllDuration.ProportionOfTimeInSight=true")
        rows.append("Report.StateAllDuration.ConditionalProportionOfTime=false")
        rows.append("")
        rows.append("Report.EventNaturalInterval.EventCount=false")
        rows.append("Report.EventNaturalInterval.Occurrence=false")
        rows.append("Report.EventNaturalInterval.Average=false")
        rows.append("Report.EventNaturalInterval.StandardDeviation=false")
        rows.append("Report.EventNaturalInterval.ConditionalNatEventCount=false")
        rows.append("Report.EventNaturalInterval.ConditionalNatRate=false")
        rows.append("Report.EventNaturalInterval.ConditionalNatIntervalOccurance=false")
        rows.append("Report.EventNaturalInterval.ConditionalNatIntervalAverage=false")
        rows.append("Report.EventNaturalInterval.ConditionalNatIntervalStandardDeviation=false")
        rows.append("Report.EventNaturalInterval.ConditionalAllEventCount=false")
        rows.append("Report.EventNaturalInterval.ConditionalAllRate=false")
        rows.append("Report.EventNaturalInterval.ConditionalAllIntervalOccurance=false")
        rows.append("Report.EventNaturalInterval.ConditionalAllIntervalAverage=false")
        rows.append("Report.EventNaturalInterval.ConditionalAllIntervalStandardDeviation=false")
        rows.append("")
        rows.append("AllCodesMutuallyExclusive=true")
        rows.append("")

        for behav, key in all_observed_behaviors:
            rows.append(f"Behavior.isModified.{key}=false")
            rows.append(f"Behavior.isSubtracted.{key}=false")
            rows.append(f"Behavior.isIgnored.{key}=false")
            rows.append(f"Behavior.isEventAnalyzed.{key}=false")
            rows.append(f"Behavior.switchesOff.{key}=")
            rows.append("")

        if faf_creation_answer == "" or faf_creation_answer == cfg.OVERWRITE:
            try:
                with open(pathlib.Path(file_name_subject).with_suffix(".faf"), "w") as f_out:
                    f_out.write("\n".join(rows))
            except Exception:
                return False, f"File FAF not created: {sys.exc_info()[1]}"

    return True, ""


def export_tabular_events(
    pj: dict, parameters: dict, obs_id: str, observation: dict, ethogram: dict, file_name: str, output_format: str
) -> Tuple[bool, str]:
    """
    export events for one observation (obs_id)

    Args:
        parameters (dict): subjects, behaviors
        obs_id (str): observation id
        observation (dict): observation
        ethogram (dict): ethogram of project
        file_name (str): file name for exporting events
        output_format (str): output for exporting events

    Returns:
        bool: result: True if OK else False
        str: error message
    """

    logging.debug(f"function: export tabular events for {obs_id}")
    logging.debug(f"parameters: {parameters}")

    interval = parameters["time"]

    start_coding, end_coding, coding_duration = observation_operations.coding_time(pj[cfg.OBSERVATIONS], [obs_id])

    if interval == cfg.TIME_EVENTS:
        min_time = start_coding
        max_time = end_coding

    if interval == cfg.TIME_FULL_OBS:
        if observation[cfg.TYPE] == cfg.MEDIA:
            max_media_duration, _ = observation_operations.media_duration(pj[cfg.OBSERVATIONS], [obs_id])
            min_time = dec("0")
            max_time = max_media_duration
            coding_duration = max_media_duration

        if observation[cfg.TYPE] in (cfg.LIVE, cfg.IMAGES):
            min_time = start_coding
            max_time = end_coding

    if interval == cfg.TIME_ARBITRARY_INTERVAL:
        min_time = parameters[cfg.START_TIME]
        max_time = parameters[cfg.END_TIME]

    logging.debug(f"min_time: {min_time}  max_time: {max_time}")

    events_with_status = project_functions.events_start_stop(ethogram, observation[cfg.EVENTS], observation[cfg.TYPE])

    # check max number of modifiers
    max_modifiers = 0
    for event in events_with_status:
        if not math.isnan(min_time) and not math.isnan(max_time):  # obs not from pictures
            if min_time <= event[cfg.EVENT_TIME_FIELD_IDX] <= max_time:
                if event[cfg.EVENT_MODIFIER_FIELD_IDX]:
                    max_modifiers = max(max_modifiers, len(event[cfg.EVENT_MODIFIER_FIELD_IDX].split("|")))
        else:
            if event[cfg.EVENT_MODIFIER_FIELD_IDX]:
                max_modifiers = max(max_modifiers, len(event[cfg.EVENT_MODIFIER_FIELD_IDX].split("|")))

    # media file number
    media_nb = util.count_media_file(observation[cfg.FILE])

    rows: list = []

    # fields and type
    fields_type: dict = {
        "Observation id": str,
        "Observation date": dt.datetime,
        "Description": str,
        "Observation duration": float,
        "Observation type": str,
        "Source": str,
        "Time offset (s)": str,
    }
    if media_nb == 1:
        fields_type["Media duration (s)"] = float
        fields_type["FPS"] = float
    else:
        fields_type["Media duration (s)"] = str
        fields_type["FPS"] = str

    # independent variables
    if cfg.INDEPENDENT_VARIABLES in observation:
        for variable in observation[cfg.INDEPENDENT_VARIABLES]:
            # TODO check variable type
            fields_type[variable] = str

    fields_type.update({"Subject": str, "Behavior": str, "Behavioral category": str})

    # modifiers
    for idx in range(max_modifiers):
        fields_type[f"Modifier #{idx + 1}"] = str

    fields_type.update(
        {
            "Behavior type": str,
            "Time": float,
            "Media file name": str,
            "Image index": float,  # add image index and image file path to header
            "Image file path": str,
            "Comment": str,
        }
    )

    # add header
    rows.append(list(fields_type.keys()))

    behavioral_category = project_functions.behavior_category(ethogram)

    duration1 = []  # in seconds
    if observation[cfg.TYPE] == cfg.MEDIA:
        try:
            for mediaFile in observation[cfg.FILE][cfg.PLAYER1]:
                duration1.append(observation[cfg.MEDIA_INFO][cfg.LENGTH][mediaFile])
        except KeyError:
            pass

    for event in events_with_status:
        if (not math.isnan(min_time)) and not (min_time <= event[cfg.EVENT_TIME_FIELD_IDX] <= max_time):
            continue
        if (
            (event[cfg.EVENT_SUBJECT_FIELD_IDX] in parameters[cfg.SELECTED_SUBJECTS])
            or (event[cfg.EVENT_SUBJECT_FIELD_IDX] == "" and cfg.NO_FOCAL_SUBJECT in parameters[cfg.SELECTED_SUBJECTS])
        ) and (event[cfg.EVENT_BEHAVIOR_FIELD_IDX] in parameters[cfg.SELECTED_BEHAVIORS]):
            fields: list = []
            fields.append(obs_id)
            fields.append(observation.get("date", "").replace("T", " "))
            fields.append(util.eol2space(observation.get(cfg.DESCRIPTION, "")))
            # total length
            fields.append(coding_duration if not coding_duration.is_nan() else cfg.NA)

            if observation[cfg.TYPE] == cfg.MEDIA:
                fields.append("Media file(s)")

                media_file_str, fps_str, media_durations_str = "", "", ""
                for player in observation[cfg.FILE]:
                    media_file_lst, fps_lst, media_durations_lst = [], [], []
                    if observation[cfg.FILE][player]:
                        for media_file in observation[cfg.FILE][player]:
                            media_file_lst.append(media_file)
                            fps_lst.append(f"{observation[cfg.MEDIA_INFO][cfg.FPS].get(media_file, cfg.NA):.3f}")
                            media_durations_lst.append(f"{observation[cfg.MEDIA_INFO][cfg.LENGTH].get(media_file, cfg.NA):.3f}")
                        if player > "1":
                            media_file_str += "|"
                            fps_str += "|"
                            media_durations_str += "|"
                        media_file_str += f"player #{player}:" + ";".join(media_file_lst)
                        fps_str += ";".join(fps_lst)
                        media_durations_str += ";".join(media_durations_lst)

                """
                # number of players
                n_players = len([x for x in observation[cfg.FILE] if observation[cfg.FILE][x]])
                media_file_str, fps_str = "", ""
                for player in observation[cfg.FILE]:
                    if observation[cfg.FILE][player]:
                        if media_file_str:
                            media_file_str += " "
                        if fps_str:
                            fps_str += " "
                        if n_players > 1:
                            media_file_str += f"player #{player}: "
                            fps_str += f"player #{player}: "
                        media_list, fps_list = [], []
                        for media_file in observation[cfg.FILE][player]:
                            media_list.append(media_file)
                            fps_list.append(f"{observation[cfg.MEDIA_INFO][cfg.FPS].get(media_file, cfg.NA):.3f}")
                        media_file_str += ";".join(media_list)
                        fps_str += ";".join(fps_list)
                """

                fields.append(media_file_str)

            elif observation[cfg.TYPE] == cfg.LIVE:
                fields.append("Live observation")
                fields.append(cfg.NA)
                media_durations_str = cfg.NA
                fps_str = cfg.NA

            elif observation[cfg.TYPE] == cfg.IMAGES:
                fields.append("From directories of images")
                dir_list = []
                for dir in observation[cfg.DIRECTORIES_LIST]:
                    dir_list.append(dir)
                fields.append(";".join(dir_list))
                media_durations_str = cfg.NA
                fps_str = cfg.NA

            else:
                fields.append("")

            # time offset
            fields.append(observation[cfg.TIME_OFFSET])

            # media duration
            fields.append(media_durations_str)

            # FPS
            fields.append(fps_str)

            # indep var
            if cfg.INDEPENDENT_VARIABLES in observation:
                for variable in observation[cfg.INDEPENDENT_VARIABLES]:
                    fields.append(observation[cfg.INDEPENDENT_VARIABLES][variable])

            fields.append(event[cfg.PJ_OBS_FIELDS[observation[cfg.TYPE]][cfg.SUBJECT]])
            fields.append(event[cfg.PJ_OBS_FIELDS[observation[cfg.TYPE]][cfg.BEHAVIOR_CODE]])

            # behavioral category
            try:
                behav_category = behavioral_category[event[cfg.PJ_OBS_FIELDS[observation[cfg.TYPE]][cfg.BEHAVIOR_CODE]]]
            except Exception:
                behav_category = ""
            fields.append(behav_category)

            # modifiers
            if max_modifiers:
                modifiers = event[cfg.PJ_OBS_FIELDS[observation[cfg.TYPE]][cfg.MODIFIER]].split("|")
                while len(modifiers) < max_modifiers:
                    modifiers.append("")

                for m in modifiers:
                    fields.append(m)

            # status (START/STOP)
            fields.append(event[-1])

            # time
            fields.append(util.convertTime(time_format=cfg.S, sec=event[cfg.EVENT_TIME_FIELD_IDX]))

            # check video file name containing the event
            if observation[cfg.TYPE] == cfg.MEDIA:
                video_file_name = observation_operations.event2media_file_name(observation, event[cfg.EVENT_TIME_FIELD_IDX])
                if video_file_name is None:
                    video_file_name = "Not found"

            elif observation[cfg.TYPE] in (cfg.LIVE, cfg.IMAGES):
                video_file_name = cfg.NA
            fields.append(video_file_name)

            # image file index
            if observation[cfg.TYPE] == cfg.IMAGES:
                fields.append(event[cfg.PJ_OBS_FIELDS[cfg.IMAGES][cfg.IMAGE_INDEX]])  # image file index
            elif observation[cfg.TYPE] == cfg.MEDIA:
                frame_idx = event_operations.read_event_field(event, cfg.MEDIA, cfg.FRAME_INDEX)
                fields.append(frame_idx)  # frame index
            elif observation[cfg.TYPE] == cfg.LIVE:
                fields.append(cfg.NA)
            else:
                fields.append("")

            # image file path
            if observation[cfg.TYPE] == cfg.IMAGES:
                fields.append(event[cfg.PJ_OBS_FIELDS[cfg.IMAGES][cfg.IMAGE_PATH]])  # image file path
            elif observation[cfg.TYPE] in (cfg.LIVE, cfg.MEDIA):
                fields.append(cfg.NA)
            else:
                fields.append("")

            # comment
            fields.append(event[cfg.PJ_OBS_FIELDS[observation[cfg.TYPE]][cfg.COMMENT]].replace(os.linesep, " "))

            rows.append(fields)

    max_len = max([len(r) for r in rows])
    data = tablib.Dataset()

    data.title = util.safe_xl_worksheet_title(obs_id, output_format)

    for row in rows:
        data.append(util.complete(row, max_len))

    r, msg = dataset_write(data, file_name, output_format, dtype=fields_type)

    return r, msg


def dataset_write(dataset: tablib.Dataset, file_name: str, output_format: str, dtype: dict = {}) -> tuple:  # -> tuple[bool, str]:
    """
    write a tablib dataset with aggregated events or tabular events to file in specified format (output_format)

    Args:
        dataset (tablib.dataset): dataset to write
        file_name (str): file name
        output_format (str): format of output
        dtype (dict): type of field

    Returns:
        bool: result. True if OK else False
        str: error message
    """

    logging.debug("function: dataset_write")

    try:
        if output_format in (cfg.PANDAS_DF_EXT, cfg.RDS_EXT):
            # build pandas dataframe from the tsv export of tablib dataset
            date_type = []
            for field_name in dtype:
                if dtype[field_name] == dt.datetime:
                    date_type.append(field_name)
            # delete data type from dtype
            for field_name in date_type:
                del dtype[field_name]

            df = pd.read_csv(
                StringIO(dataset.export(cfg.TSV_EXT)),
                sep="\t",
                dtype=dtype,
                parse_dates=date_type,
            )

            if output_format == cfg.PANDAS_DF_EXT:
                df.to_pickle(file_name)

            if flag_pyreadr_loaded and output_format == cfg.RDS_EXT:
                pyreadr.write_rds(file_name, df)

            return True, ""

        if output_format in (cfg.CSV_EXT, cfg.TSV_EXT, cfg.HTML_EXT):
            with open(file_name, "wb") as f:
                f.write(str.encode(dataset.export(output_format)))
            return True, ""

        if output_format in (cfg.ODS_EXT, cfg.XLS_EXT, cfg.XLSX_EXT):
            dataset.title = util.safe_xl_worksheet_title(dataset.title, output_format)

            with open(file_name, "wb") as f:
                f.write(dataset.export(output_format))
            return True, ""

        return False, f"Format {output_format} not found"

    except Exception:
        return False, str(sys.exc_info()[1])


def export_aggregated_events(pj: dict, parameters: dict, obsId: str, force_number_modifiers: int = 0) -> Tuple[tablib.Dataset, int]:
    """
    export aggregated events of one observation

    Args:
        pj (dict): BORIS project
        parameters (dict): subjects, behaviors
        obsId (str): observation id
        force_number_modifiers (int): force the number of modifiers to return

    Returns:
        tablib.Dataset:
        int: Maximum number of modifiers

    """
    logging.debug(f"function: export aggregated events {obsId} parameters: {parameters} ")

    observation = pj[cfg.OBSERVATIONS][obsId]
    interval = parameters["time"]

    data = tablib.Dataset()

    start_coding, end_coding, coding_duration = observation_operations.coding_time(pj[cfg.OBSERVATIONS], [obsId])
    if start_coding is None and end_coding is None:  # no events
        return data, 0

    if interval == cfg.TIME_EVENTS:
        min_time = float(start_coding)
        max_time = float(end_coding)

    if interval == cfg.TIME_FULL_OBS:
        if observation[cfg.TYPE] == cfg.MEDIA:
            max_media_duration, _ = observation_operations.media_duration(pj[cfg.OBSERVATIONS], [obsId])
            min_time = float(0)
            max_time = float(max_media_duration)
            coding_duration = max_media_duration
        if observation[cfg.TYPE] in (cfg.LIVE, cfg.IMAGES):
            min_time = float(start_coding)
            max_time = float(end_coding)

    if interval == cfg.TIME_ARBITRARY_INTERVAL:
        min_time = float(parameters[cfg.START_TIME])
        max_time = float(parameters[cfg.END_TIME])

    logging.debug(f"min_time: {min_time}  max_time: {max_time}")

    # obs description
    obs_description = util.eol2space(observation[cfg.DESCRIPTION])

    """
    obs_length = observation_operations.observation_total_length(pj[cfg.OBSERVATIONS][obsId])
    logging.debug(f"obs_length: {obs_length}")
    """

    _, _, connector = db_functions.load_aggregated_events_in_db(
        pj, parameters[cfg.SELECTED_SUBJECTS], [obsId], parameters[cfg.SELECTED_BEHAVIORS]
    )
    if connector is None:
        logging.critical("error when loading aggregated events in DB")
        return data, 0

    cursor = connector.cursor()

    # adapt start and stop to the selected time interval
    if not math.isnan(min_time) and not math.isnan(max_time):  # obs with timestamp
        # delete events outside time interval
        cursor.execute(
            "DELETE FROM aggregated_events WHERE observation = ? AND (start < ? AND stop < ?) OR (start > ? AND stop > ?)",
            (
                obsId,
                min_time,
                min_time,
                max_time,
                max_time,
            ),
        )

        cursor.execute(
            "UPDATE aggregated_events SET start = ? WHERE observation = ? AND start < ? AND stop BETWEEN ? AND ?",
            (
                min_time,
                obsId,
                min_time,
                min_time,
                max_time,
            ),
        )
        cursor.execute(
            "UPDATE aggregated_events SET stop = ? WHERE observation = ? AND stop > ? AND start BETWEEN ? AND ?",
            (
                max_time,
                obsId,
                max_time,
                min_time,
                max_time,
            ),
        )

        cursor.execute(
            "UPDATE aggregated_events SET start = ?, stop = ? WHERE observation = ? AND start < ? AND stop > ?",
            (
                min_time,
                max_time,
                obsId,
                min_time,
                max_time,
            ),
        )

    behavioral_category = project_functions.behavior_category(pj[cfg.ETHOGRAM])

    cursor.execute("SELECT DISTINCT modifiers FROM aggregated_events")

    if not force_number_modifiers:
        max_modifiers: int = 0
        for row in cursor.fetchall():
            if row["modifiers"]:
                max_modifiers = max(max_modifiers, row["modifiers"].count("|") + 1)
    else:
        max_modifiers = force_number_modifiers

    for subject in parameters[cfg.SELECTED_SUBJECTS]:
        # calculate observation duration by subject (by obs)
        cursor.execute(("SELECT SUM(stop - start) AS duration FROM aggregated_events WHERE subject = ? "), (subject,))
        duration_by_subject_by_obs = cursor.fetchone()["duration"]
        if duration_by_subject_by_obs is not None:
            duration_by_subject_by_obs = round(duration_by_subject_by_obs, 3)

        for behavior in parameters[cfg.SELECTED_BEHAVIORS]:
            cursor.execute(
                "SELECT DISTINCT modifiers FROM aggregated_events WHERE subject=? AND behavior=? ORDER BY modifiers",
                (
                    subject,
                    behavior,
                ),
            )

            rows_distinct_modifiers = list(x[0] for x in cursor.fetchall())

            for distinct_modifiers in rows_distinct_modifiers:
                cursor.execute(
                    (
                        "SELECT start, stop, type, modifiers, comment, comment_stop, "
                        "image_index_start, image_index_stop, image_path_start, image_path_stop "
                        "FROM aggregated_events "
                        "WHERE subject = ? AND behavior = ? AND modifiers = ? ORDER BY start, image_index_start"
                    ),
                    (subject, behavior, distinct_modifiers),
                )

                for row in cursor.fetchall():
                    media_file_name = cfg.NA

                    if observation[cfg.TYPE] == cfg.MEDIA:
                        observation_type = "Media file"

                        media_file_name = observation_operations.event2media_file_name(observation, row["start"])
                        if media_file_name is None:
                            media_file_name = "Not found"

                        media_file_str, fps_str, media_durations_str = "", "", ""

                        for player in observation[cfg.FILE]:
                            media_file_lst, fps_lst, media_durations_lst = [], [], []
                            if observation[cfg.FILE][player]:
                                for media_file in observation[cfg.FILE][player]:
                                    media_file_lst.append(media_file)
                                    fps_lst.append(f"{observation[cfg.MEDIA_INFO][cfg.FPS].get(media_file, cfg.NA):.3f}")
                                    media_durations_lst.append(f"{observation[cfg.MEDIA_INFO][cfg.LENGTH].get(media_file, cfg.NA):.3f}")
                                if player > "1":
                                    media_file_str += "|"
                                    fps_str += "|"
                                    media_durations_str += "|"
                                media_file_str += f"player #{player}:" + ";".join(media_file_lst)
                                fps_str += ";".join(fps_lst)
                                media_durations_str += ";".join(media_durations_lst)

                    if observation[cfg.TYPE] == cfg.LIVE:
                        observation_type = "Live observation"
                        media_file_str = cfg.NA
                        fps_str = cfg.NA
                        media_durations_str = cfg.NA

                    if observation[cfg.TYPE] == cfg.IMAGES:
                        observation_type = "From pictures"
                        media_file_str = ""
                        for dir in observation[cfg.DIRECTORIES_LIST]:
                            media_file_str += f"{dir}; "
                        fps_str = cfg.NA
                        # TODO: number of pictures in each directory
                        media_durations_str = cfg.NA

                    row_data = []

                    row_data.extend(
                        [
                            obsId,
                            observation["date"].replace("T", " "),
                            obs_description,
                            observation_type,
                            media_file_str,
                            pj[cfg.OBSERVATIONS][obsId][cfg.TIME_OFFSET],
                            f"{coding_duration:.3f}" if not coding_duration.is_nan() else cfg.NA,
                            media_durations_str,
                            fps_str,
                        ]
                    )

                    # independent variables
                    if cfg.INDEPENDENT_VARIABLES in pj:
                        for idx_var in util.sorted_keys(pj[cfg.INDEPENDENT_VARIABLES]):
                            if pj[cfg.INDEPENDENT_VARIABLES][idx_var]["label"] in observation[cfg.INDEPENDENT_VARIABLES]:
                                var_value = observation[cfg.INDEPENDENT_VARIABLES][pj[cfg.INDEPENDENT_VARIABLES][idx_var]["label"]]
                                if pj[cfg.INDEPENDENT_VARIABLES][idx_var]["type"] == "timestamp":
                                    var_value = var_value.replace("T", " ")

                                row_data.append(var_value)
                            else:
                                row_data.append("")

                    row_data.extend(
                        [
                            subject,
                            duration_by_subject_by_obs,
                            behavior,
                            behavioral_category[behavior] if behavioral_category[behavior] else "Not defined",
                        ]
                    )

                    # modifiers
                    if max_modifiers:
                        modifiers = row["modifiers"].split("|")
                        while len(modifiers) < max_modifiers:
                            modifiers.append("")
                        for modifier in modifiers:
                            row_data.append(modifier)

                    if row["type"] == cfg.POINT:
                        row_data.extend(
                            [
                                cfg.POINT,
                                f"{row['start']:.3f}" if row["start"] is not None else cfg.NA,  # start
                                f"{row['stop']:.3f}" if row["stop"] is not None else cfg.NA,  # stop
                                cfg.NA,  # duration
                                media_file_name,  # Media file name
                            ]
                        )

                    if row["type"] == cfg.STATE:
                        row_data.extend(
                            [
                                cfg.STATE,
                                f"{row['start']:.3f}" if row["start"] is not None else cfg.NA,
                                f"{row['stop']:.3f}" if row["stop"] is not None else cfg.NA,
                                # duration
                                f"{row['stop'] - row['start']:.3f}" if (row["stop"] is not None) and (row["start"] is not None) else cfg.NA,
                                media_file_name,  # Media file name
                            ]
                        )

                    row_data.extend(
                        [
                            row["image_index_start"],
                            row["image_index_stop"],
                            row["image_path_start"],
                            row["image_path_stop"],
                            row["comment"],
                            row["comment_stop"] if (row["type"] == cfg.STATE) else "",
                        ]
                    )

                    data.append(row_data)

    return data, max_modifiers


def events_to_behavioral_sequences(pj, obs_id: str, subj: str, parameters: dict, behav_seq_separator: str) -> str:
    """
    return the behavioral sequence (behavioral string) for subject in obs_id

    Args:
        pj (dict): project
        obs_id (str): observation id
        subj (str): subject
        parameters (dict): parameters
        behav_seq_separator (str): separator of behviors in behavioral sequences

    Returns:
        str: behavioral string for selected subject in selected observation
    """

    out: str = ""
    current_states: list = []
    # add status (POINT, START, STOP) to event
    events_with_status = project_functions.events_start_stop(
        pj[cfg.ETHOGRAM], pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS], pj[cfg.OBSERVATIONS][obs_id][cfg.TYPE]
    )

    for event in events_with_status:
        # check if event in selected behaviors

        if event[cfg.EVENT_BEHAVIOR_FIELD_IDX] not in parameters[cfg.SELECTED_BEHAVIORS]:
            continue

        if event[cfg.EVENT_SUBJECT_FIELD_IDX] == subj or (subj == cfg.NO_FOCAL_SUBJECT and event[cfg.EVENT_SUBJECT_FIELD_IDX] == ""):
            # if event[cfg.EVENT_STATUS_FIELD_IDX] == cfg.POINT:
            if event[-1] == cfg.POINT:  # status is last element
                if current_states:
                    out += "+".join(current_states) + "+" + event[cfg.EVENT_BEHAVIOR_FIELD_IDX]
                else:
                    out += event[cfg.EVENT_BEHAVIOR_FIELD_IDX]

                if parameters[cfg.INCLUDE_MODIFIERS]:
                    out += "&" + event[cfg.EVENT_MODIFIER_FIELD_IDX].replace("|", "+")

                out += behav_seq_separator

            # if event[cfg.EVENT_STATUS_FIELD_IDX] == cfg.START:
            if event[-1] == cfg.START:  # status is last element
                if parameters[cfg.INCLUDE_MODIFIERS]:
                    current_states.append(
                        (
                            f"{event[cfg.EVENT_BEHAVIOR_FIELD_IDX]}"
                            f"{'&' if event[cfg.EVENT_MODIFIER_FIELD_IDX] else ''}"
                            f"{event[cfg.EVENT_MODIFIER_FIELD_IDX].replace('|', ';')}"
                        )
                    )
                else:
                    current_states.append(event[cfg.EVENT_BEHAVIOR_FIELD_IDX])

                out += "+".join(sorted(current_states))

                out += behav_seq_separator

            # if event[cfg.EVENT_STATUS_FIELD_IDX] == cfg.STOP:
            if event[-1] == cfg.STOP:
                if parameters[cfg.INCLUDE_MODIFIERS]:
                    behav_modif = (
                        f"{event[cfg.EVENT_BEHAVIOR_FIELD_IDX]}"
                        f"{'&' if event[cfg.EVENT_MODIFIER_FIELD_IDX] else ''}"
                        f"{event[cfg.EVENT_MODIFIER_FIELD_IDX].replace('|', ';')}"
                    )
                else:
                    behav_modif = event[cfg.EVENT_BEHAVIOR_FIELD_IDX]
                if behav_modif in current_states:
                    current_states.remove(behav_modif)

                if current_states:
                    out += "+".join(sorted(current_states))

                    out += behav_seq_separator

    # remove last separator (if separator not empty)
    if behav_seq_separator:
        out = out[0 : -len(behav_seq_separator)]

    return out


def events_to_behavioral_sequences_all_subj(pj, obs_id: str, subjects_list: list, parameters: dict, behav_seq_separator: str) -> str:
    """
    return the behavioral sequences for all selected subjects in obs_id

    Args:
        pj (dict): project
        obs_id (str): observation id
        subjects_list (list): list of subjects
        parameters (dict): parameters
        behav_seq_separator (str): separator of behviors in behavioral sequences

    Returns:
        str: behavioral sequences for all selected subjects in selected observation
    """

    out = ""
    current_states = {i: [] for i in subjects_list}
    events_with_status = project_functions.events_start_stop(
        pj[cfg.ETHOGRAM], pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS], pj[cfg.OBSERVATIONS][obs_id][cfg.TYPE]
    )

    for event in events_with_status:
        # check if event in selected behaviors
        if event[cfg.EVENT_BEHAVIOR_FIELD_IDX] not in parameters[cfg.SELECTED_BEHAVIORS]:
            continue

        if (event[cfg.EVENT_SUBJECT_FIELD_IDX] in subjects_list) or (
            event[cfg.EVENT_SUBJECT_FIELD_IDX] == "" and cfg.NO_FOCAL_SUBJECT in subjects_list
        ):
            subject = event[cfg.EVENT_SUBJECT_FIELD_IDX] if event[cfg.EVENT_SUBJECT_FIELD_IDX] else cfg.NO_FOCAL_SUBJECT

            if event[-1] == cfg.POINT:
                if current_states[subject]:
                    out += f"[{subject}]" + "+".join(current_states[subject]) + "+" + event[cfg.EVENT_BEHAVIOR_FIELD_IDX]
                else:
                    out += f"[{subject}]" + event[cfg.EVENT_BEHAVIOR_FIELD_IDX]

                if parameters[cfg.INCLUDE_MODIFIERS]:
                    out += "&" + event[cfg.EVENT_MODIFIER_FIELD_IDX].replace("|", "+")

                out += behav_seq_separator

            if event[-1] == cfg.START:
                if parameters[cfg.INCLUDE_MODIFIERS]:
                    current_states[subject].append(
                        (
                            f"{event[cfg.EVENT_BEHAVIOR_FIELD_IDX]}"
                            f"{'&' if event[cfg.EVENT_MODIFIER_FIELD_IDX] else ''}"
                            f"{event[cfg.EVENT_MODIFIER_FIELD_IDX].replace('|', ';')}"
                        )
                    )
                else:
                    current_states[subject].append(event[cfg.EVENT_BEHAVIOR_FIELD_IDX])

                out += f"[{subject}]" + "+".join(sorted(current_states[subject]))

                out += behav_seq_separator

            if event[-1] == cfg.STOP:
                if parameters[cfg.INCLUDE_MODIFIERS]:
                    behav_modif = (
                        f"{event[cfg.EVENT_BEHAVIOR_FIELD_IDX]}"
                        f"{'&' if event[cfg.EVENT_MODIFIER_FIELD_IDX] else ''}"
                        f"{event[cfg.EVENT_MODIFIER_FIELD_IDX].replace('|', ';')}"
                    )
                else:
                    behav_modif = event[cfg.EVENT_BEHAVIOR_FIELD_IDX]
                if behav_modif in current_states[subject]:
                    current_states[subject].remove(behav_modif)

                if current_states[subject]:
                    out += f"[{subject}]" + "+".join(sorted(current_states[subject]))

                    out += behav_seq_separator

    # remove last separator (if separator not empty)
    if behav_seq_separator:
        out = out[0 : -len(behav_seq_separator)]

    return out


def events_to_timed_behavioral_sequences(
    pj: dict, obs_id: str, subject: str, parameters: dict, precision: float, behav_seq_separator: str
) -> str:
    """
    return the behavioral string for subject in obsId

    Args:
        pj (dict): project
        obs_id (str): observation id
        subj (str): subject
        parameters (dict): parameters
        precision (float): time value for scan sample
        behav_seq_separator (str): separator of behviors in behavioral sequences

    Returns:
        str: behavioral string for selected subject in selected observation
    """

    out: str = ""

    state_behaviors_codes = util.state_behavior_codes(pj[cfg.ETHOGRAM])
    delta = dec(str(round(precision, 3)))
    out = ""
    t = dec("0.000")

    current = []
    while t < pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS][-1][0]:
        """
        if out:
            out += behav_seq_separator
        """
        csbs = util.get_current_states_modifiers_by_subject(
            state_behaviors_codes,
            pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS],
            {"": {"name": subject}},
            t,
            include_modifiers=False,
        )[""]
        if csbs:
            if current:
                if csbs == current[-1]:
                    current.append("+".join(csbs))
                else:
                    out.append(current)
                    current = [csbs]
            else:
                current = [csbs]

        t += delta

    return out


def observation_to_behavioral_sequences(pj, selected_observations, parameters, behaviors_separator, separated_subjects, timed, file_name):
    try:
        with open(file_name, "w", encoding="utf-8") as out_file:
            for obs_id in selected_observations:
                # observation id
                out_file.write("\n" + f"# observation id: {obs_id}" + "\n")
                # observation description
                descr = pj[cfg.OBSERVATIONS][obs_id]["description"]
                if "\r\n" in descr:
                    descr = descr.replace("\r\n", "\n# ")
                elif "\r" in descr:
                    descr = descr.replace("\r", "\n# ")
                out_file.write(f"# observation description: {descr}\n\n")
                # media file name
                if pj[cfg.OBSERVATIONS][obs_id][cfg.TYPE] == cfg.MEDIA:
                    out_file.write(f"# Observation type: Media file{os.linesep}")
                    media_file_str = ""

                    for player in pj[cfg.OBSERVATIONS][obs_id][cfg.FILE]:
                        if pj[cfg.OBSERVATIONS][obs_id][cfg.FILE][player]:
                            media_file_str += f"player #{player}: "
                            # fps_str += f"player #{player}: "
                            for media_file in pj[cfg.OBSERVATIONS][obs_id][cfg.FILE][player]:
                                media_file_str += f"{media_file}; "
                                # fps_str += f"{pj[cfg.OBSERVATIONS][obs_id][cfg.MEDIA_INFO][cfg.FPS].get(media_file, cfg.NA):.3f}; "

                    out_file.write(f"# Media file path: {media_file_str}\n\n")
                if pj[cfg.OBSERVATIONS][obs_id][cfg.TYPE] == cfg.LIVE:
                    out_file.write(f"# Observation type: Live observation{os.linesep}{os.linesep}")

                if pj[cfg.OBSERVATIONS][obs_id][cfg.TYPE] == cfg.IMAGES:
                    out_file.write(f"# Observation type: From pictures{os.linesep}{os.linesep}")

                # independent variables
                if cfg.INDEPENDENT_VARIABLES in pj[cfg.OBSERVATIONS][obs_id]:
                    out_file.write("# Independent variables\n")

                    for variable in pj[cfg.OBSERVATIONS][obs_id][cfg.INDEPENDENT_VARIABLES]:
                        out_file.write(f"# {variable}: {pj[cfg.OBSERVATIONS][obs_id][cfg.INDEPENDENT_VARIABLES][variable]}\n")
                out_file.write("\n")

                # one sequence for all subjects
                if not separated_subjects:
                    out = events_to_behavioral_sequences_all_subj(
                        pj, obs_id, parameters[cfg.SELECTED_SUBJECTS], parameters, behaviors_separator
                    )
                    if out:
                        out_file.write(out + "\n")

                # one sequence by subject
                if separated_subjects:
                    # selected subjects
                    for subject in parameters[cfg.SELECTED_SUBJECTS]:
                        out_file.write(f"\n# {subject if subject else cfg.NO_FOCAL_SUBJECT}:\n")

                        if not timed:
                            out = events_to_behavioral_sequences(pj, obs_id, subject, parameters, behaviors_separator)

                        if timed:
                            out = events_to_timed_behavioral_sequences(pj, obs_id, subject, parameters, 0.001, behaviors_separator)

                        if out:
                            out_file.write(out + "\n")

            return True, ""

    except Exception:
        error_type, error_file_name, error_lineno = util.error_info(sys.exc_info())
        return False, f"{error_type} {error_file_name} {error_lineno}"



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2024 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

import sys
import json
import urllib.error
import urllib.parse
import urllib.request


from PyQt5.QtWidgets import QMessageBox, QTableWidgetItem, QFileDialog, QInputDialog, QLineEdit

from . import dialog
from . import config as cfg


def pb_code_help_clicked(self):
    """
    help for writing converters
    """
    msg = QMessageBox()
    msg.setIcon(QMessageBox.Information)
    msg.setWindowTitle("Help for writing converters")

    msg.setText(
        (
            "A converter is a function that will convert a time value from external data into seconds.<br>"
            "A time value like 00:23:59 must be converted into seconds before to be plotted synchronously with your media.<br>"
            "For this you can use BORIS native converters or write your own converter.<br>"
            'A converter must be written using the <a href="www.python.org">Python3</a> language.<br>'
        )
    )

    # msg.setInformativeText("This is additional information")

    msg.setStandardButtons(QMessageBox.Ok)
    msg.exec_()


def add_converter(self):
    """
    Add a new converter
    """

    for w in [
        self.le_converter_name,
        self.le_converter_description,
        self.pteCode,
        self.pb_save_converter,
        self.pb_cancel_converter,
    ]:
        w.setEnabled(True)
    # disable buttons
    for w in [
        self.pb_add_converter,
        self.pb_modify_converter,
        self.pb_delete_converter,
        self.pb_load_from_file,
        self.pb_load_from_repo,
        self.tw_converters,
    ]:
        w.setEnabled(False)


def modify_converter(self):
    """
    Modifiy the selected converter
    """

    if not self.tw_converters.selectedIndexes():
        QMessageBox.warning(self, cfg.programName, "Select a converter in the table")
        return

    for w in [
        self.le_converter_name,
        self.le_converter_description,
        self.pteCode,
        self.pb_save_converter,
        self.pb_cancel_converter,
    ]:
        w.setEnabled(True)

    # disable buttons
    for w in [
        self.pb_add_converter,
        self.pb_modify_converter,
        self.pb_delete_converter,
        self.pb_load_from_file,
        self.pb_load_from_repo,
        self.tw_converters,
    ]:
        w.setEnabled(False)

    self.le_converter_name.setText(self.tw_converters.item(self.tw_converters.selectedIndexes()[0].row(), 0).text())
    self.le_converter_description.setText(self.tw_converters.item(self.tw_converters.selectedIndexes()[0].row(), 1).text())
    self.pteCode.setPlainText(self.tw_converters.item(self.tw_converters.selectedIndexes()[0].row(), 2).text().replace("@", "\n"))

    self.row_in_modification = self.tw_converters.selectedIndexes()[0].row()


def code_2_func(self, name: str, code: str):
    """
    convert code to function

    Args:
        name (str): function name
        code (str): Python code

    Returns:
        str: string containing Python function
    """

    function = f"def {name}(INPUT):\n"
    function += """    INPUT = INPUT.decode("utf-8") if isinstance(INPUT, bytes) else INPUT\n"""
    function += "\n".join(["    " + row for row in code.split("\n")])
    function += "\n    return OUTPUT"

    return function


def save_converter(self):
    """Save converter"""

    # check if name
    self.le_converter_name.setText(self.le_converter_name.text().strip())
    if not self.le_converter_name.text():
        QMessageBox.critical(self, "BORIS", "The converter must have a name")
        return

    if not self.le_converter_name.text().replace("_", "a").isalnum():
        QMessageBox.critical(self, "BORIS", "Forbidden characters are used in converter name.<br>Use a..z, A..Z, 0..9 _")
        return

    # test code with exec
    code: str = self.pteCode.toPlainText()
    if not code:
        QMessageBox.critical(self, "BORIS", "The converter must have Python code")
        return

    function = code_2_func(self, name=self.le_converter_name.text(), code=code)

    try:
        exec(function)
    except Exception:
        QMessageBox.critical(self, "BORIS", f"The code produces an error:<br><b>{sys.exc_info()[1]}</b>")
        return

    if self.row_in_modification == -1:
        self.tw_converters.setRowCount(self.tw_converters.rowCount() + 1)
        row = self.tw_converters.rowCount() - 1
    else:
        row = self.row_in_modification

    self.tw_converters.setItem(row, 0, QTableWidgetItem(self.le_converter_name.text()))
    self.tw_converters.setItem(row, 1, QTableWidgetItem(self.le_converter_description.text()))
    self.tw_converters.setItem(row, 2, QTableWidgetItem(self.pteCode.toPlainText().replace("\n", "@")))

    self.row_in_modification = -1

    for w in [self.le_converter_name, self.le_converter_description, self.pteCode]:
        w.setEnabled(False)
        w.clear()
    self.pb_save_converter.setEnabled(False)
    self.pb_cancel_converter.setEnabled(False)
    self.tw_converters.setEnabled(True)

    self.flag_modified = True

    # enable buttons
    for w in [
        self.pb_add_converter,
        self.pb_modify_converter,
        self.pb_delete_converter,
        self.pb_load_from_file,
        self.pb_load_from_repo,
        self.tw_converters,
    ]:
        w.setEnabled(True)


def cancel_converter(self):
    """Cancel converter"""

    for w in [self.le_converter_name, self.le_converter_description, self.pteCode]:
        w.setEnabled(False)
        w.clear()
    self.pb_save_converter.setEnabled(False)
    self.pb_cancel_converter.setEnabled(False)

    # enable buttons
    for w in [
        self.pb_add_converter,
        self.pb_modify_converter,
        self.pb_delete_converter,
        self.pb_load_from_file,
        self.pb_load_from_repo,
        self.tw_converters,
    ]:
        w.setEnabled(True)


def delete_converter(self):
    """
    Delete selected converter
    """

    if self.tw_converters.selectedIndexes():
        if dialog.MessageDialog("BORIS", "Confirm converter deletion", [cfg.CANCEL, cfg.OK]) == cfg.OK:
            self.tw_converters.removeRow(self.tw_converters.selectedIndexes()[0].row())
    else:
        QMessageBox.warning(self, cfg.programName, "Select a converter in the table")


def load_converters_from_file_repo(self, mode: str):
    """
    Load converters from file (JSON) or BORIS remote repository
    Args:
        mode (str): string "repo" or "file"
    """

    converters_from_file = {}
    if mode == "file":
        fn = QFileDialog(self).getOpenFileName(self, "Load converters from file", "", "All files (*)")
        file_name = fn[0] if type(fn) is tuple else fn

        if file_name:
            with open(file_name, "r") as f_in:
                try:
                    converters_from_file = json.loads(f_in.read())["BORIS converters"]
                except Exception:
                    QMessageBox.critical(self, cfg.programName, "This file does not contain converters...")
                    return

    if mode == "repo":
        converters_repo_URL = "https://www.boris.unito.it/static/converters.json"
        try:
            converters_from_repo = urllib.request.urlopen(converters_repo_URL).read().strip().decode("utf-8")
        except Exception:
            QMessageBox.critical(self, cfg.programName, "An error occured during retrieving converters from BORIS remote repository")
            return

        try:
            converters_from_file = eval(converters_from_repo)["BORIS converters"]
        except Exception:
            QMessageBox.critical(self, cfg.programName, "An error occured during retrieving converters from BORIS remote repository")
            return

    if converters_from_file:
        diag_choose_conv = dialog.ChooseObservationsToImport("Choose the converters to load:", sorted(list(converters_from_file.keys())))

        if diag_choose_conv.exec_():
            selected_converters = diag_choose_conv.get_selected_observations()
            if selected_converters:
                # extract converter names from table
                converter_names = []
                for row in range(self.tw_converters.rowCount()):
                    converter_names.append(self.tw_converters.item(row, 0).text())

                for converter in selected_converters:
                    converter_name = converter

                    if converter in converter_names:
                        while True:
                            text, ok = QInputDialog.getText(
                                self,
                                "Converter conflict",
                                "The converter already exists<br>Rename it:",
                                QLineEdit.Normal,
                                converter,
                            )
                            if not ok:
                                break
                            if text in converter_names:
                                QMessageBox.critical(self, cfg.programName, "This name already exists in converters")

                            if not text.replace("_", "a").isalnum():
                                QMessageBox.critical(
                                    self,
                                    cfg.programName,
                                    "This name contains forbidden character(s).<br>Use a..z, A..Z, 0..9 _",
                                )

                            if text != converter and text not in converter_names and text.replace("_", "a").isalnum():
                                break

                        if ok:
                            converter_name = text
                        else:
                            continue
                    # test if code does not produce error
                    function = code_2_func(self, name=converter_name, code=converters_from_file[converter]["code"])

                    try:
                        exec(function)
                    except Exception:
                        QMessageBox.critical(
                            self,
                            "BORIS",
                            (f"The code of {converter_name} converter produces an error: " f"<br><b>{sys.exc_info()[1]}</b>"),
                        )

                    self.tw_converters.setRowCount(self.tw_converters.rowCount() + 1)
                    self.tw_converters.setItem(self.tw_converters.rowCount() - 1, 0, QTableWidgetItem(converter_name))
                    self.tw_converters.setItem(
                        self.tw_converters.rowCount() - 1,
                        1,
                        QTableWidgetItem(converters_from_file[converter]["description"]),
                    )
                    self.tw_converters.setItem(
                        self.tw_converters.rowCount() - 1,
                        2,
                        QTableWidgetItem(converters_from_file[converter]["code"].replace("\n", "@")),
                    )

                    self.flag_modified = True

            [self.tw_converters.resizeColumnToContents(idx) for idx in [0, 1]]



# -*- coding: utf-8 -*-
"""
Utilities for processing SASS and images from default and custom palette.
"""

# Standard library imports
import logging
import sys

# Local imports
from qdarkstyle import PACKAGE_PATH, QRC_FILE_SUFFIX
from qdarkstyle.utils.images import (compile_qrc_file, create_images,
                                     create_palette_image, generate_qrc_file)
from qdarkstyle.utils.scss import create_qss

_logger = logging.getLogger(__name__)


def process_palette(palette, compile_for='qtpy'):
    """Process palette class to create a new palette file/folders.

    It generates all files below, in this order:
        - Palette files (svg/.png) under docs/images/[palette_id]
        - Image files (.png) under [palette_id]/rc folder.
        - QRC file in [palette_id]/[palette_id]style.qrc (C++).
        - SCSS variables in [palette_id]/_variables.scss file.
        - QSS file in [palette_id]/[palette_id]style.qss.
        - Compiled QRC file in [palette_id]/[palette_id]style_rc.py

    TODO:
        - Must generalize to create custom palettes and folder paths.
        - Must create/copy all files under [palette_id], such as main.scss,
            __init__.py, palette.py.
        - Add option or avoid adding the palette under docs for custom palettes.

    Args:
        palette (Palette): Palette.
        compile_for (list, optional): Prefix used in resources.
            Defaults to 'qtpy'. Possible values are 'qtpy', 'pyqtgraph',
            'pyqt', 'pyqt5', 'pyqt6',
            'pyside', 'pyside2', 'pyside6',
            'qt', 'qt5', qt6, 'all'.
    """

    if palette is None:
        _logger.error("Please pass a palette class in order to create its "
                      "associated images")
        sys.exit(1)

    if palette.ID is None:
        _logger.error("A QDarkStyle palette requires an ID!")
        sys.exit(1)

    id_ = palette.ID
    print(f"-- PROCESSING THEME: {id_}")

    # TODO: delete/remove all files and folders to ensure that old files
    # are not used

    print(f"-- GENERATING PALETTE IMAGE FOR: {id_}")
    create_palette_image(palette=palette)

    print(f"-- GENERATING IMAGE FILES (.svg > .png) FOR: {id_}")
    create_images(palette=palette)

    print(f"-- GENERATING QRC FILE FOR: {id_}")
    generate_qrc_file(palette=palette)

    print(f"-- GENERATING QSS FILE (.scss > .qss) FOR: {id_}")
    create_qss(palette=palette)

    print(f"-- CONVERTING RESOURCE FILE (. qrc > _rc.py/.rcc) FOR: {id_}")
    compile_qrc_file(compile_for=compile_for, palette=palette)



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2024 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

import logging
import os
import pathlib
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), ".")))
import qdarkstyle

from . import dialog
from . import gui_utilities
from . import menu_options
from . import config as cfg
from . import config_file

from .preferences_ui import Ui_prefDialog

from PyQt5.QtWidgets import QDialog, QFileDialog


class Preferences(QDialog, Ui_prefDialog):
    def __init__(self, parent=None):
        super().__init__()
        self.setupUi(self)

        self.pbBrowseFFmpegCacheDir.clicked.connect(self.browseFFmpegCacheDir)

        self.pb_reset_behav_colors.clicked.connect(self.reset_behav_colors)
        self.pb_reset_category_colors.clicked.connect(self.reset_category_colors)

        self.pb_refresh.clicked.connect(self.refresh_preferences)
        self.pbOK.clicked.connect(self.accept)
        self.pbCancel.clicked.connect(self.reject)

        self.flag_refresh = False
        self.flag_reset_frames_memory = False

    def refresh_preferences(self):
        """
        allow user to delete the config file (.boris)
        """
        if (
            dialog.MessageDialog(
                "BORIS",
                ("Refresh will re-initialize " "all your preferences and close BORIS"),
                [cfg.CANCEL, "Refresh preferences"],
            )
            == "Refresh preferences"
        ):
            self.flag_refresh = True
            self.accept()

    def browseFFmpegCacheDir(self):
        """
        allow user select a cache dir for ffmpeg images
        """
        FFmpegCacheDir = QFileDialog().getExistingDirectory(
            self,
            "Select a directory",
            os.path.expanduser("~"),
            options=QFileDialog().ShowDirsOnly,
        )
        if FFmpegCacheDir:
            self.leFFmpegCacheDir.setText(FFmpegCacheDir)

    def reset_behav_colors(self):
        """
        reset behavior colors to default
        """
        self.te_behav_colors.setPlainText("\n".join(cfg.BEHAVIORS_PLOT_COLORS))

        logging.debug("reset behaviors colors to default")

    def reset_category_colors(self):
        """
        reset category colors to default
        """
        self.te_category_colors.setPlainText("\n".join(cfg.CATEGORY_COLORS_LIST))

        logging.debug("reset category colors to default")


def preferences(self):
    """
    show preferences window
    """

    preferencesWindow = Preferences()
    preferencesWindow.tabWidget.setCurrentIndex(0)

    if self.timeFormat == cfg.S:
        preferencesWindow.cbTimeFormat.setCurrentIndex(0)

    if self.timeFormat == cfg.HHMMSS:
        preferencesWindow.cbTimeFormat.setCurrentIndex(1)

    preferencesWindow.sbffSpeed.setValue(self.fast)
    preferencesWindow.cb_adapt_fast_jump.setChecked(self.config_param.get(cfg.ADAPT_FAST_JUMP, False))
    preferencesWindow.sbRepositionTimeOffset.setValue(self.repositioningTimeOffset)
    preferencesWindow.sbSpeedStep.setValue(self.play_rate_step)
    # automatic backup
    preferencesWindow.sbAutomaticBackup.setValue(self.automaticBackup)
    # separator for behavioural strings
    preferencesWindow.leSeparator.setText(self.behav_seq_separator)
    # close same event indep of modifiers
    preferencesWindow.cbCloseSameEvent.setChecked(self.close_the_same_current_event)
    # confirm sound
    preferencesWindow.cbConfirmSound.setChecked(self.confirmSound)
    # beep every
    preferencesWindow.sbBeepEvery.setValue(self.beep_every)
    # alert no focal subject
    preferencesWindow.cbAlertNoFocalSubject.setChecked(self.alertNoFocalSubject)
    # tracking cursor above event
    preferencesWindow.cbTrackingCursorAboveEvent.setChecked(self.trackingCursorAboveEvent)
    # check for new version
    preferencesWindow.cbCheckForNewVersion.setChecked(self.checkForNewVersion)
    # display subtitles
    preferencesWindow.cb_display_subtitles.setChecked(self.config_param[cfg.DISPLAY_SUBTITLES])
    # pause before add event
    preferencesWindow.cb_pause_before_addevent.setChecked(self.pause_before_addevent)
    # MPV hwdec
    preferencesWindow.cb_hwdec.clear()
    preferencesWindow.cb_hwdec.addItems(cfg.MPV_HWDEC_OPTIONS)
    try:
        preferencesWindow.cb_hwdec.setCurrentIndex(
            cfg.MPV_HWDEC_OPTIONS.index(self.config_param.get(cfg.MPV_HWDEC, cfg.MPV_HWDEC_DEFAULT_VALUE))
        )
    except Exception:
        preferencesWindow.cb_hwdec.setCurrentIndex(cfg.MPV_HWDEC_DEFAULT_VALUE)

    # PROJET FILE INDENTATION
    preferencesWindow.combo_project_file_indentation.clear()
    preferencesWindow.combo_project_file_indentation.addItems(cfg.PROJECT_FILE_INDENTATION_COMBO_OPTIONS)
    try:
        preferencesWindow.combo_project_file_indentation.setCurrentIndex(
            cfg.PROJECT_FILE_INDENTATION_OPTIONS.index(
                self.config_param.get(
                    cfg.PROJECT_FILE_INDENTATION,
                    cfg.PROJECT_FILE_INDENTATION_DEFAULT_VALUE,
                )
            )
        )
    except Exception:
        preferencesWindow.combo_project_file_indentation.setCurrentText(
            cfg.PROJECT_FILE_INDENTATION_COMBO_OPTIONS[
                cfg.PROJECT_FILE_INDENTATION_OPTIONS.index(cfg.PROJECT_FILE_INDENTATION_DEFAULT_VALUE)
            ]
        )

    # FFmpeg for frame by frame mode
    preferencesWindow.lbFFmpegPath.setText(f"FFmpeg path: {self.ffmpeg_bin}")
    preferencesWindow.leFFmpegCacheDir.setText(self.ffmpeg_cache_dir)

    # spectrogram
    preferencesWindow.cbSpectrogramColorMap.clear()
    preferencesWindow.cbSpectrogramColorMap.addItems(cfg.SPECTROGRAM_COLOR_MAPS)
    try:
        preferencesWindow.cbSpectrogramColorMap.setCurrentIndex(cfg.SPECTROGRAM_COLOR_MAPS.index(self.spectrogram_color_map))
    except Exception:
        preferencesWindow.cbSpectrogramColorMap.setCurrentIndex(cfg.SPECTROGRAM_COLOR_MAPS.index(cfg.SPECTROGRAM_DEFAULT_COLOR_MAP))

    try:
        preferencesWindow.cbSpectrogramColorMap.setCurrentIndex(cfg.SPECTROGRAM_COLOR_MAPS.index(self.spectrogram_color_map))
    except Exception:
        preferencesWindow.cbSpectrogramColorMap.setCurrentIndex(cfg.SPECTROGRAM_COLOR_MAPS.index(cfg.SPECTROGRAM_DEFAULT_COLOR_MAP))

    try:
        preferencesWindow.sb_time_interval.setValue(self.spectrogram_time_interval)
    except Exception:
        preferencesWindow.sb_time_interval.setValue(cfg.SPECTROGRAM_DEFAULT_TIME_INTERVAL)

    # behavior colors
    if not self.plot_colors:
        self.plot_colors = cfg.BEHAVIORS_PLOT_COLORS
    preferencesWindow.te_behav_colors.setPlainText("\n".join(self.plot_colors))

    # category colors
    if not self.behav_category_colors:
        self.behav_category_colors = cfg.CATEGORY_COLORS_LIST
    preferencesWindow.te_category_colors.setPlainText("\n".join(self.behav_category_colors))

    # interface
    preferencesWindow.sb_toolbar_icon_size.setValue(self.config_param.get(cfg.TOOLBAR_ICON_SIZE, cfg.DEFAULT_TOOLBAR_ICON_SIZE_VALUE))
    preferencesWindow.cb_darkmode.setChecked(self.config_param.get(cfg.DARK_MODE, cfg.DARK_MODE_DEFAULT_VALUE))

    self.config_param[cfg.DARK_MODE] = preferencesWindow.cb_darkmode.isChecked()

    gui_utilities.restore_geometry(preferencesWindow, "preferences", (700, 500))

    if preferencesWindow.exec_():
        gui_utilities.save_geometry(preferencesWindow, "preferences")

        if preferencesWindow.flag_refresh:
            # refresh preferences remove the config file

            logging.debug("flag refresh ")

            self.config_param["refresh_preferences"] = True
            self.close()
            # check if refresh canceled for not saved project
            if "refresh_preferences" in self.config_param:
                if (pathlib.Path.home() / ".boris").exists():
                    os.remove(pathlib.Path.home() / ".boris")
                sys.exit()

        if preferencesWindow.cbTimeFormat.currentIndex() == 0:
            self.timeFormat = cfg.S

        if preferencesWindow.cbTimeFormat.currentIndex() == 1:
            self.timeFormat = cfg.HHMMSS

        self.fast = preferencesWindow.sbffSpeed.value()

        self.config_param[cfg.ADAPT_FAST_JUMP] = preferencesWindow.cb_adapt_fast_jump.isChecked()

        self.repositioningTimeOffset = preferencesWindow.sbRepositionTimeOffset.value()

        self.play_rate_step = preferencesWindow.sbSpeedStep.value()

        self.automaticBackup = preferencesWindow.sbAutomaticBackup.value()
        if self.automaticBackup:
            self.automaticBackupTimer.start(self.automaticBackup * 60000)
        else:
            self.automaticBackupTimer.stop()

        self.behav_seq_separator = preferencesWindow.leSeparator.text()

        self.close_the_same_current_event = preferencesWindow.cbCloseSameEvent.isChecked()

        self.confirmSound = preferencesWindow.cbConfirmSound.isChecked()

        self.beep_every = preferencesWindow.sbBeepEvery.value()

        self.alertNoFocalSubject = preferencesWindow.cbAlertNoFocalSubject.isChecked()

        self.trackingCursorAboveEvent = preferencesWindow.cbTrackingCursorAboveEvent.isChecked()

        self.checkForNewVersion = preferencesWindow.cbCheckForNewVersion.isChecked()

        self.config_param[cfg.DISPLAY_SUBTITLES] = preferencesWindow.cb_display_subtitles.isChecked()

        self.pause_before_addevent = preferencesWindow.cb_pause_before_addevent.isChecked()

        # MPV hwdec
        self.config_param[cfg.MPV_HWDEC] = cfg.MPV_HWDEC_OPTIONS[preferencesWindow.cb_hwdec.currentIndex()]

        # project file indentation
        self.config_param[cfg.PROJECT_FILE_INDENTATION] = cfg.PROJECT_FILE_INDENTATION_OPTIONS[
            preferencesWindow.combo_project_file_indentation.currentIndex()
        ]

        if self.observationId:
            self.load_tw_events(self.observationId)
            self.display_statusbar_info(self.observationId)

        self.ffmpeg_cache_dir = preferencesWindow.leFFmpegCacheDir.text()

        # spectrogram
        self.spectrogram_color_map = preferencesWindow.cbSpectrogramColorMap.currentText()
        # self.spectrogramHeight = preferencesWindow.sbSpectrogramHeight.value()
        self.spectrogram_time_interval = preferencesWindow.sb_time_interval.value()

        # behav colors
        self.plot_colors = preferencesWindow.te_behav_colors.toPlainText().split()
        # category colors
        self.behav_category_colors = preferencesWindow.te_category_colors.toPlainText().split()

        # interface
        self.config_param[cfg.TOOLBAR_ICON_SIZE] = preferencesWindow.sb_toolbar_icon_size.value()
        self.config_param[cfg.DARK_MODE] = preferencesWindow.cb_darkmode.isChecked()
        if self.config_param[cfg.DARK_MODE]:
            self.setStyleSheet(qdarkstyle.load_stylesheet(qt_api="pyqt5"))
        else:
            self.setStyleSheet("")

        menu_options.update_menu(self)

        config_file.save(self)



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2024 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

import binascii
import json
import os

from PyQt5.QtCore import (
    Qt,
    pyqtSignal,
    QPoint,
    QByteArray,
    QBuffer,
    QIODevice,
    QLineF,
)
from PyQt5.QtGui import (
    QColor,
    QBrush,
    QMouseEvent,
    QPixmap,
    QIcon,
    QPen,
    QPolygon,
    QPolygonF,
)

from PyQt5.QtWidgets import (
    QGraphicsPolygonItem,
    QGraphicsEllipseItem,
    QGraphicsPixmapItem,
    QGraphicsLineItem,
    QAction,
    QMainWindow,
    QGraphicsView,
    QPushButton,
    QLabel,
    QHBoxLayout,
    QLineEdit,
    QSlider,
    QGraphicsScene,
    QWidget,
    QColorDialog,
    QVBoxLayout,
    QMessageBox,
    QInputDialog,
    QFileDialog,
    QApplication,
)

from . import config as cfg
from . import dialog
from . import utilities as util

designColor = QColor(255, 0, 0, 128)  # red opacity: 50%
penWidth = 0
penStyle = Qt.NoPen
selectedBrush = QBrush()
selectedBrush.setStyle(Qt.SolidPattern)
selectedBrush.setColor(QColor(255, 255, 0, 255))


class ModifiersMapCreatorWindow(QMainWindow):
    closed = pyqtSignal()

    class View(QGraphicsView):
        """
        class for handling mousepress event in QGraphicsView
        """

        mousePress = pyqtSignal(QMouseEvent)

        def mousePressEvent(self, event):
            self.mousePress.emit(event)

        _start = 0
        elList, points = [], []

        def __init__(self, parent):
            QGraphicsView.__init__(self, parent)
            self.setBackgroundBrush(QColor(128, 128, 128))
            self.setScene(QGraphicsScene(self))
            self.scene().update()

    bitmapFileName, mapName, fileName = "", "", ""
    flagNewArea, flagMapChanged = False, False
    areasList, polygonsList2 = {}, {}
    areaColor = QColor("lime")

    def __init__(self):
        super(ModifiersMapCreatorWindow, self).__init__()

        self.pixmap = QPixmap()
        self.closedPolygon = None
        self.selectedPolygon = None

        self.setWindowTitle("BORIS - Modifiers map creator")

        self.newMapAction = QAction(QIcon(), "&New modifiers map", self)
        self.newMapAction.setShortcut("Ctrl+N")
        self.newMapAction.setStatusTip("Create a new modifiers map")
        self.newMapAction.triggered.connect(self.newMap)

        self.openMapAction = QAction(QIcon(), "&Open modifiers map", self)
        self.openMapAction.setShortcut("Ctrl+O")
        self.openMapAction.setStatusTip("Open a modifiers map")
        self.openMapAction.triggered.connect(self.openMap)

        self.saveMapAction = QAction(QIcon(), "&Save modifiers map", self)
        self.saveMapAction.setShortcut("Ctrl+S")
        self.saveMapAction.setStatusTip("Save modifiers map")
        self.saveMapAction.setEnabled(False)
        self.saveMapAction.triggered.connect(self.saveMap_clicked)

        self.saveAsMapAction = QAction(QIcon(), "Save modifiers map as", self)
        self.saveAsMapAction.setStatusTip("Save modifiers map as")
        self.saveAsMapAction.setEnabled(False)
        self.saveAsMapAction.triggered.connect(self.saveAsMap_clicked)

        self.mapNameAction = QAction(QIcon(), "&Modifiers map name", self)
        self.mapNameAction.setShortcut("Ctrl+M")
        self.mapNameAction.setStatusTip("Change modifiers map name")
        self.mapNameAction.setEnabled(False)
        self.mapNameAction.triggered.connect(self.mapName_clicked)

        self.exitAction = QAction(QIcon(), "&Close", self)
        self.exitAction.setStatusTip("Close modifiers map creator")
        self.exitAction.triggered.connect(self.close)

        menubar = self.menuBar()
        fileMenu = menubar.addMenu("&Modifiers Map creator")
        fileMenu.addAction(self.newMapAction)
        fileMenu.addAction(self.openMapAction)
        fileMenu.addAction(self.saveMapAction)
        fileMenu.addAction(self.saveAsMapAction)
        fileMenu.addSeparator()
        fileMenu.addAction(self.mapNameAction)
        fileMenu.addSeparator()
        fileMenu.addAction(self.exitAction)

        self.view = self.View(self)
        self.view.mousePress.connect(self.viewMousePressEvent)

        self.btLoad = QPushButton("Load bitmap", self)
        self.btLoad.clicked.connect(self.loadBitmap)
        self.btLoad.setVisible(False)

        self.btNewArea = QPushButton("New modifier", self)
        self.btNewArea.clicked.connect(self.newArea)
        self.btNewArea.setVisible(False)

        self.hlayout = QHBoxLayout()

        self.lb = QLabel("Modifier")
        self.lb.setVisible(False)
        self.hlayout.addWidget(self.lb)

        self.leAreaCode = QLineEdit(self)
        self.leAreaCode.setVisible(False)
        self.hlayout.addWidget(self.leAreaCode)

        self.btColor = QPushButton()
        self.btColor.clicked.connect(self.chooseColor)
        self.btColor.setVisible(False)
        self.btColor.setStyleSheet("QWidget {{background-color:{}}}".format(self.areaColor.name()))
        self.hlayout.addWidget(self.btColor)

        self.slAlpha = QSlider(Qt.Horizontal)
        self.slAlpha.setRange(20, 100)
        self.slAlpha.setValue(50)
        self.slAlpha.valueChanged.connect(self.slAlpha_changed)
        self.slAlpha.setVisible(False)
        self.hlayout.addWidget(self.slAlpha)

        self.slAlpha_changed(50)
        """
        self.btCancelMap = QPushButton("Cancel modifiers map", self)
        self.btCancelMap.clicked.connect(self.cancelMap)
        self.btCancelMap.setVisible(False)
        """

        layout = QVBoxLayout()
        layout.addWidget(self.view)
        layout.addWidget(self.btLoad)

        hlayout2 = QHBoxLayout()
        hlayout2.addWidget(self.btNewArea)

        self.btSaveArea = QPushButton("Save modifier", self)
        self.btSaveArea.clicked.connect(self.saveArea)
        self.btSaveArea.setVisible(False)
        hlayout2.addWidget(self.btSaveArea)

        self.btCancelAreaCreation = QPushButton("Cancel new modifier", self)
        self.btCancelAreaCreation.clicked.connect(self.cancelAreaCreation)
        self.btCancelAreaCreation.setVisible(False)
        hlayout2.addWidget(self.btCancelAreaCreation)

        self.btDeleteArea = QPushButton("Delete selected modifier", self)
        self.btDeleteArea.clicked.connect(self.deleteArea)
        self.btDeleteArea.setVisible(True)
        self.btDeleteArea.setEnabled(False)
        hlayout2.addWidget(self.btDeleteArea)

        layout.addLayout(hlayout2)
        layout.addLayout(self.hlayout)
        """layout.addWidget(self.btCancelMap)"""

        main_widget = QWidget(self)
        main_widget.setLayout(layout)
        self.setCentralWidget(main_widget)

        self.statusBar().showMessage("")

    def slAlpha_changed(self, val):
        """
        opacity slider value changed
        """

        self.btColor.setText("Opacity: {} %".format(val))
        self.areaColor.setAlpha(int(val / 100 * 255))

        if self.selectedPolygon:
            self.selectedPolygon.setBrush(self.areaColor)
            self.areasList[self.leAreaCode.text()]["color"] = self.areaColor.rgba()

        if self.closedPolygon:
            self.closedPolygon.setBrush(self.areaColor)

    def chooseColor(self):
        """
        area color button clicked
        """
        cd = QColorDialog()
        cd.setWindowFlags(Qt.WindowStaysOnTopHint)
        cd.setOptions(QColorDialog.ShowAlphaChannel | QColorDialog.DontUseNativeDialog)

        if cd.exec_():
            self.areaColor = cd.currentColor()
            self.btColor.setStyleSheet(f"QWidget {{background-color:{self.areaColor.name()}}}")

        if self.selectedPolygon:
            self.selectedPolygon.setBrush(self.areaColor)
            self.areasList[self.leAreaCode.text()]["color"] = self.areaColor.rgba()

        if self.closedPolygon:
            self.closedPolygon.setBrush(self.areaColor)

    def closeEvent(self, event):
        if self.flagMapChanged:
            response = dialog.MessageDialog(
                "BORIS - Modifiers map creator",
                "What to do about the current unsaved modifiers coding map?",
                ["Save", "Discard", "Cancel"],
            )

            if response == "Save":
                if not self.saveMap_clicked():
                    event.ignore()

            if response == "Cancel":
                event.ignore()
                return

        self.closed.emit()
        event.accept()

    def viewMousePressEvent(self, event):
        """
        check if area selected with mouse
        """

        if not self.bitmapFileName:
            return

        self.btDeleteArea.setEnabled(False)

        test = self.view.mapToScene(event.pos()).toPoint()

        if test.x() < 0 or test.y() < 0 or test.x() > self.pixmap.size().width() or test.y() > self.pixmap.size().height():
            return

        if not self.flagNewArea:  # test clicked point for areas
            txt = ""

            # reset selected polygon to default pen
            if self.selectedPolygon:
                self.selectedPolygon.setPen(QPen(designColor, penWidth, penStyle, Qt.RoundCap, Qt.RoundJoin))
                self.selectedPolygon = None
                self.selectedPolygonMemBrush = None

            for areaCode in self.polygonsList2:
                if self.polygonsList2[areaCode].contains(test):
                    if txt:
                        txt += ","

                    txt += areaCode
                    self.selectedPolygon = self.polygonsList2[areaCode]
                    self.selectedPolygonAreaCode = areaCode

                    self.selectedPolygonMemBrush = self.selectedPolygon.brush()

                    self.selectedPolygon.setPen(
                        QPen(
                            QColor(255, 0, 0, 255),
                            2,
                            Qt.SolidLine,
                            Qt.RoundCap,
                            Qt.RoundJoin,
                        )
                    )

                    self.leAreaCode.setText(areaCode)
                    self.leAreaCode.setVisible(True)
                    self.btDeleteArea.setEnabled(True)

                    self.areaColor = self.selectedPolygon.brush().color()
                    self.btColor.setStyleSheet("QWidget {{background-color:{}}}".format(self.selectedPolygon.brush().color().name()))
                    self.btColor.setVisible(True)

                    self.slAlpha.setValue(int(self.selectedPolygon.brush().color().alpha() / 255 * 100))
                    self.slAlpha.setVisible(True)

                    break

            if txt:
                self.statusBar().showMessage("Modifier{}: {}".format("s" if "," in txt else "", txt))
            else:
                self.statusBar().showMessage("")

            if not self.selectedPolygon:
                self.leAreaCode.setVisible(False)
                self.btColor.setVisible(False)
                self.slAlpha.setVisible(False)
            return

        # delete last line item
        if (event.buttons() & Qt.RightButton) and not self.closedPolygon:
            if self.view.points:
                self.view.points = self.view.points[0:-1]

            if self.view.points:
                self.view._start = QPoint(self.view.points[-1][0], self.view.points[-1][1])
            else:
                self.view._start = None

            # remove graphical elements
            if self.view.elList:
                self.view.scene().removeItem(self.view.elList[-1])
                self.view.elList = self.view.elList[0:-1]

        # add line item
        if event.buttons() == Qt.LeftButton and not self.closedPolygon:
            if self.view._start:
                end = test

                # test is polygon is crossed
                if len(self.view.points) >= 3:
                    for idx, point in enumerate(self.view.points[:-2]):
                        if util.intersection(
                            self.view.points[idx],
                            self.view.points[idx + 1],
                            self.view.points[-1],
                            (int(end.x()), int(end.y())),
                        ):
                            QMessageBox.critical(self, "", "The polygon edges can not be intersected")
                            return

                # test if polygon closed (dist min 10 px)
                if abs(end.x() - self.view.points[0][0]) < 10 and abs(end.y() - self.view.points[0][1]) < 10:
                    line = QGraphicsLineItem(
                        QLineF(
                            self.view._start,
                            QPoint(self.view.points[0][0], self.view.points[0][1]),
                        )
                    )
                    line.setPen(
                        QPen(
                            designColor,
                            penWidth,
                            Qt.SolidLine,
                            Qt.RoundCap,
                            Qt.RoundJoin,
                        )
                    )

                    self.view.scene().addItem(line)
                    self.view.elList.append(line)

                    self.statusBar().showMessage("Area completed")

                    # create polygon
                    newPolygon = QPolygonF()
                    for p in self.view.points:
                        newPolygon.append(QPoint(p[0], p[1]))

                    # draw polygon a red polygon
                    self.closedPolygon = QGraphicsPolygonItem(newPolygon)

                    self.closedPolygon.setPen(QPen(designColor, penWidth, penStyle, Qt.RoundCap, Qt.RoundJoin))

                    self.closedPolygon.setBrush(self.areaColor)

                    self.view.scene().addItem(self.closedPolygon)

                    return

                self.view.points.append((int(end.x()), int(end.y())))

                line = QGraphicsLineItem(QLineF(self.view._start, end))

                line.setPen(QPen(designColor, 2, Qt.SolidLine, Qt.RoundCap, Qt.RoundJoin))

                self.view.scene().addItem(line)
                self.view.elList.append(line)

                self.view._start = test

            else:  # first point
                self.view._start = test

                ellipse = QGraphicsEllipseItem(self.view._start.x(), self.view._start.y(), 3, 3)
                ellipse.setPen(QPen(designColor, 1, Qt.SolidLine, Qt.RoundCap, Qt.RoundJoin))

                brush = QBrush()
                brush.setStyle(Qt.SolidPattern)
                brush.setColor(designColor)
                ellipse.setBrush(brush)

                self.view.scene().addItem(ellipse)
                self.view.elList.append(ellipse)

                self.view.points.append((self.view._start.x(), self.view._start.y()))

    def mapName_clicked(self):
        """
        change map name
        """
        text, ok = QInputDialog.getText(
            self,
            "Modifiers map name",
            "Enter a name for the modifiers map",
            QLineEdit.Normal,
            self.mapName,
        )
        if ok:
            self.mapName = text
            self.setWindowTitle("{} - Modifiers map creator tool - {}".format(cfg.programName, self.mapName))

    def newMap(self):
        """
        create a new map
        """

        if self.flagMapChanged:
            response = dialog.MessageDialog(
                cfg.programName + " - Modifiers map creator",
                "What to do about the current unsaved coding map?",
                ["Save", "Discard", "Cancel"],
            )

            if response == "Save":
                if not self.saveMap_clicked():
                    return

            if response == "Cancel":
                return

        self.cancelMap()

        text, ok = QInputDialog.getText(self, "Map name", "Enter a name for the new map")
        if ok:
            self.mapName = text
        else:
            return

        if self.mapName == "":
            QMessageBox.critical(self, "", "You must define a name for the new map")
            return

        if self.mapName in ["areas", "bitmap"]:
            QMessageBox.critical(self, "", "This name is not allowed")
            return

        self.setWindowTitle(cfg.programName + " - Map creator tool - " + self.mapName)

        self.btLoad.setVisible(True)
        """self.btCancelMap.setVisible(True)"""

        self.statusBar().showMessage('Click "Load bitmap" button to select and load a bitmap into the viewer')

    def openMap(self):
        """
        load bitmap from data
        show it in view scene
        """
        if self.flagMapChanged:
            response = dialog.MessageDialog(
                cfg.programName + " - Map creator",
                "What to do about the current unsaved coding map?",
                ["Save", "Discard", "Cancel"],
            )

            if response == "Save":
                if not self.saveMap_clicked():
                    return

            if response == "Cancel":
                return

        fn = QFileDialog().getOpenFileName(
            self,
            "Open a coding map",
            "",
            "BORIS coding map (*.boris_map);;All files (*)",
        )
        fileName = fn[0] if type(fn) is tuple else fn

        if fileName:
            try:
                self.codingMap = json.loads(open(fileName, "r").read())
            except Exception:
                QMessageBox.critical(
                    self,
                    cfg.programName,
                    "The file {} seems not a behaviors coding map...".format(fileName),
                )
                return

            self.cancelMap()

            self.mapName = self.codingMap["name"]

            self.setWindowTitle(cfg.programName + " - Map creator tool - " + self.mapName)

            self.bitmapFileName = True

            self.fileName = fileName

            self.areasList = self.codingMap["areas"]  # dictionary of dictionaries
            bitmapContent = binascii.a2b_base64(self.codingMap["bitmap"])

            self.pixmap.loadFromData(bitmapContent)

            self.btDeleteArea.setEnabled(False)

            self.view.setSceneRect(0, 0, self.pixmap.size().width(), self.pixmap.size().height())
            pixItem = QGraphicsPixmapItem(self.pixmap)
            pixItem.setPos(0, 0)
            self.view.scene().addItem(pixItem)

            for areaCode in self.areasList:
                points = self.areasList[areaCode]["geometry"]

                newPolygon = QPolygonF()
                for p in points:
                    newPolygon.append(QPoint(p[0], p[1]))

                clr = QColor()
                clr.setRgba(self.areasList[areaCode]["color"])

                # draw polygon
                polygon = QGraphicsPolygonItem()

                polygon.setPolygon(newPolygon)

                polygon.setPen(QPen(clr, penWidth, penStyle, Qt.RoundCap, Qt.RoundJoin))

                polygon.setBrush(QBrush(clr, Qt.SolidPattern))

                self.view.scene().addItem(polygon)
                self.polygonsList2[areaCode] = polygon

            self.btNewArea.setVisible(True)

            self.btLoad.setVisible(False)

            self.saveMapAction.setEnabled(True)
            self.saveAsMapAction.setEnabled(True)
            self.mapNameAction.setEnabled(True)
            self.statusBar().showMessage('Click "New area" to create a new area')
        else:
            self.statusBar().showMessage("No file", 5000)

    def saveMap(self):
        if self.fileName:
            # create dict with map name key
            mapDict = {"name": self.mapName}

            # add areas
            mapDict["areas"] = self.areasList

            import io

            # Save QPixmap to QByteArray via QBuffer.
            byte_array = QByteArray()
            buffer = QBuffer(byte_array)
            buffer.open(QIODevice.WriteOnly)
            self.pixmap.save(buffer, "PNG")

            string_io = io.BytesIO(byte_array)

            string_io.seek(0)

            # add bitmap
            mapDict["bitmap"] = binascii.b2a_base64(string_io.read()).decode("utf-8")

            with open(self.fileName, "w") as outfile:
                outfile.write(json.dumps(mapDict))

            self.flagMapChanged = False

            return True
        else:
            return False

    def saveAsMap_clicked(self):
        filters = "Modifiers map (*.boris_map);;All files (*)"

        fn = QFileDialog(self).getSaveFileName(self, "Save modifiers map as", "", filters)
        if type(fn) is tuple:
            self.fileName, _ = fn
        else:
            self.fileName = fn

        if self.fileName:
            if os.path.splitext(self.fileName)[1] != ".boris_map":
                self.fileName += ".boris_map"
            self.saveMap()

    def saveMap_clicked(self):
        if not self.fileName:
            fn = QFileDialog(self).getSaveFileName(
                self,
                "Save modifiers map",
                self.mapName + ".boris_map",
                "BORIS MAP (*.boris_map);;All files (*)",
            )
            if type(fn) is tuple:
                self.fileName, _ = fn
            else:
                self.fileName = fn

            if self.fileName and os.path.splitext(self.fileName)[1] != ".boris_map":
                self.fileName += ".boris_map"

        if self.fileName:
            return self.saveMap()

        return False

    def newArea(self):
        if not self.bitmapFileName:
            QMessageBox.critical(self, cfg.programName, "A bitmap must be loaded before to define areas")
            return

        if self.selectedPolygon:
            self.selectedPolygon.setPen(QPen(designColor, penWidth, penStyle, Qt.RoundCap, Qt.RoundJoin))
            self.selectedPolygon = None

        self.flagNewArea = True
        self.btSaveArea.setVisible(True)
        self.btCancelAreaCreation.setVisible(True)
        self.btNewArea.setVisible(False)
        self.lb.setVisible(True)
        self.leAreaCode.clear()
        self.leAreaCode.setVisible(True)
        self.btColor.setVisible(True)
        self.slAlpha.setVisible(True)
        self.btDeleteArea.setVisible(False)

        self.statusBar().showMessage(
            "Select the vertices of the area for this modifier with the mouse (right click will cancel the last point)"
        )

    def saveArea(self):
        if not self.closedPolygon:
            QMessageBox.critical(
                self,
                cfg.programName,
                "You must close your area before saving it.\nThe last vertex must correspond to the first one.",
            )

        if len(self.view.points) < 3:
            QMessageBox.critical(self, cfg.programName, "You must define a closed area")
            return

        # check if no area code
        if not self.leAreaCode.text():
            QMessageBox.critical(self, cfg.programName, "You must define a code for the new modifier")
            return

        # check if not allowed character
        for c in "|,()":
            if c in self.leAreaCode.text():
                QMessageBox.critical(
                    self,
                    cfg.programName,
                    "The modifier contains a character that is not allowed <b>()|,</b>.",
                )
                return

        # check if area code already used

        if self.leAreaCode.text() in self.areasList:
            QMessageBox.critical(self, cfg.programName, "The modifier is already in use")
            return

        # create polygon
        newPolygon = QPolygon()
        for p in self.view.points:
            newPolygon.append(QPoint(p[0], p[1]))

        self.areasList[self.leAreaCode.text()] = {
            "geometry": self.view.points,
            "color": self.areaColor.rgba(),
        }

        # remove all lines
        for line in self.view.elList:
            self.view.scene().removeItem(line)

        # draw polygon
        self.closedPolygon.setBrush(QBrush(self.areaColor, Qt.SolidPattern))
        self.polygonsList2[self.leAreaCode.text()] = self.closedPolygon
        self.closedPolygon = None
        self.view._start = 0
        self.view.points = []
        self.view.elList = []
        self.flagNewArea = False
        self.closedPolygon = None

        self.btSaveArea.setVisible(False)
        self.btCancelAreaCreation.setVisible(False)
        self.lb.setVisible(False)
        self.leAreaCode.setVisible(False)
        self.btColor.setVisible(False)
        self.slAlpha.setVisible(False)
        self.btDeleteArea.setVisible(True)
        self.btNewArea.setVisible(True)

        self.leAreaCode.setText("")

        self.flagMapChanged = True
        self.statusBar().showMessage("New modifier saved", 5000)

    def cancelAreaCreation(self):
        if self.closedPolygon:
            self.view.scene().removeItem(self.closedPolygon)
            self.closedPolygon = None

        # remove all lines
        for line in self.view.elList:
            self.view.scene().removeItem(line)

        self.view.elList = []

        self.view._start = 0
        self.view.points = []
        self.flagNewArea = False
        self.btCancelAreaCreation.setVisible(False)
        self.btDeleteArea.setVisible(True)
        self.btSaveArea.setVisible(False)
        self.lb.setVisible(False)

        self.btColor.setVisible(False)
        self.slAlpha.setVisible(False)
        self.btNewArea.setVisible(True)

        self.leAreaCode.setVisible(False)
        self.leAreaCode.setText("")

    def deleteArea(self):
        """
        remove selected area from map
        """

        if self.selectedPolygon:
            self.view.scene().removeItem(self.selectedPolygon)
            self.view.scene().removeItem(self.polygonsList2[self.selectedPolygonAreaCode])

            del self.polygonsList2[self.selectedPolygonAreaCode]
            del self.areasList[self.selectedPolygonAreaCode]

            self.flagMapChanged = True

        self.view.elList = []

        self.view._start = 0
        self.view.points = []
        self.flagNewArea = False
        self.btSaveArea.setVisible(False)
        self.lb.setVisible(False)

        self.btColor.setVisible(False)
        self.slAlpha.setVisible(False)
        self.btNewArea.setVisible(True)

        self.leAreaCode.setVisible(False)
        self.leAreaCode.setText("")

    def cancelMap(self):
        """
        remove current map
        """
        self.flagNewArea = False
        self.areasList = {}
        self.polygonsList2 = {}
        self.view.scene().clear()
        self.btLoad.setVisible(False)
        self.btDeleteArea.setVisible(False)
        self.btNewArea.setVisible(False)
        self.saveMapAction.setEnabled(False)
        self.saveAsMapAction.setEnabled(False)
        self.mapNameAction.setEnabled(False)
        self.statusBar().showMessage("")
        self.flagMapChanged = False

    def loadBitmap(self):
        """
        load bitmap as background for coding map
        resize bitmap to 512 px if bigger
        """

        maxSize = 512

        fn = QFileDialog().getOpenFileName(self, "Load bitmap", "", "bitmap files (*.png *.jpg);;All files (*)")
        fileName = fn[0] if type(fn) is tuple else fn

        if fileName:
            self.bitmapFileName = fileName

            self.pixmap.load(self.bitmapFileName)

            if self.pixmap.size().width() > maxSize or self.pixmap.size().height() > maxSize:
                self.pixmap = self.pixmap.scaled(maxSize, maxSize, Qt.KeepAspectRatio)
                QMessageBox.information(
                    self,
                    cfg.programName,
                    "The bitmap was resized to %d x %d pixels\nThe original file was not modified"
                    % (self.pixmap.size().width(), self.pixmap.size().height()),
                )

            # scale image
            # pixmap = pixmap.scaled (256, 256, Qt.KeepAspectRatio)

            self.view.setSceneRect(0, 0, self.pixmap.size().width(), self.pixmap.size().height())
            pixitem = QGraphicsPixmapItem(self.pixmap)
            pixitem.setPos(0, 0)
            self.view.scene().addItem(pixitem)

            self.btNewArea.setVisible(True)

            self.btLoad.setVisible(False)
            self.saveMapAction.setEnabled(True)
            self.saveAsMapAction.setEnabled(True)
            self.mapNameAction.setEnabled(True)

            self.statusBar().showMessage("""Click "New modifier" to create a new modifier""")

            self.flagMapChanged = True


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)
    window = ModifiersMapCreatorWindow()
    window.resize(640, 640)
    window.show()
    sys.exit(app.exec_())



import operator
from functools import partial

from .const import inf


def iterate(interval, step, *, base=None, reverse=False):
    """
    Iterate on the (discrete) values of given interval.

    This function returns a (lazy) iterator over the values of given interval,
    starting from its lower bound and ending on its upper bound (if interval is not open).
    Each returned value merely corresponds to lower + i * step, where "step" defines
    the step between consecutive values.
    It also accepts a callable that is used to compute the next possible
    value based on the current one.

    When a non-atomic interval is provided, this function chains the iterators obtained
    by calling itself on the underlying atomic intervals.

    The values returned by the iterator can be aligned with a base value with the "base" parameter.
    This parameter must be a callable that accepts the lower bound of the (atomic) interval as
    input, and returns the first value that needs to be considered for the iteration.
    By default, the identity function is used. If reverse=True, then the upper bound will be
    passed instead of the lower one.

    :param interval: an interval.
    :param step: step between values, or a callable that returns the next value.
    :param base: a callable that accepts a bound and returns an initial value to consider.
    :param reverse: set to True for descending order.
    :return: a lazy iterator.
    """
    if base is None:
        base = (lambda x: x)

    exclude = operator.lt if not reverse else operator.gt
    include = operator.le if not reverse else operator.ge
    step = step if callable(step) else partial(operator.add, step)

    value = base(interval.lower if not reverse else interval.upper)
    if (value == -inf and not reverse) or (value == inf and reverse):
        raise ValueError('Cannot start iteration with infinity.')

    for i in interval if not reverse else reversed(interval):
        value = base(i.lower if not reverse else i.upper)

        while exclude(value, i):
            value = step(value)

        while include(value, i):
            yield value
            value = step(value)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'edit_event.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(413, 488)
        self.verticalLayout = QtWidgets.QVBoxLayout(Form)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.label = QtWidgets.QLabel(Form)
        self.label.setObjectName("label")
        self.horizontalLayout_3.addWidget(self.label)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.pb_set_to_current_time = QtWidgets.QPushButton(Form)
        self.pb_set_to_current_time.setObjectName("pb_set_to_current_time")
        self.horizontalLayout_2.addWidget(self.pb_set_to_current_time)
        self.cb_set_time_na = QtWidgets.QCheckBox(Form)
        self.cb_set_time_na.setObjectName("cb_set_time_na")
        self.horizontalLayout_2.addWidget(self.cb_set_time_na)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem)
        self.horizontalLayout_3.addLayout(self.horizontalLayout_2)
        self.verticalLayout.addLayout(self.horizontalLayout_3)
        self.horizontalLayout_7 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        self.lb_image_idx = QtWidgets.QLabel(Form)
        self.lb_image_idx.setObjectName("lb_image_idx")
        self.horizontalLayout_7.addWidget(self.lb_image_idx)
        self.sb_image_idx = QtWidgets.QSpinBox(Form)
        self.sb_image_idx.setMaximum(10000000)
        self.sb_image_idx.setObjectName("sb_image_idx")
        self.horizontalLayout_7.addWidget(self.sb_image_idx)
        self.pb_set_to_current_image_index = QtWidgets.QPushButton(Form)
        self.pb_set_to_current_image_index.setObjectName("pb_set_to_current_image_index")
        self.horizontalLayout_7.addWidget(self.pb_set_to_current_image_index)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_7.addItem(spacerItem1)
        self.verticalLayout.addLayout(self.horizontalLayout_7)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.lbSubject = QtWidgets.QLabel(Form)
        self.lbSubject.setObjectName("lbSubject")
        self.horizontalLayout_4.addWidget(self.lbSubject)
        self.cobSubject = QtWidgets.QComboBox(Form)
        self.cobSubject.setObjectName("cobSubject")
        self.horizontalLayout_4.addWidget(self.cobSubject)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_4.addItem(spacerItem2)
        self.verticalLayout.addLayout(self.horizontalLayout_4)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.label_2 = QtWidgets.QLabel(Form)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout_5.addWidget(self.label_2)
        self.cobCode = QtWidgets.QComboBox(Form)
        self.cobCode.setObjectName("cobCode")
        self.horizontalLayout_5.addWidget(self.cobCode)
        spacerItem3 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_5.addItem(spacerItem3)
        self.verticalLayout.addLayout(self.horizontalLayout_5)
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.lb_frame_idx = QtWidgets.QLabel(Form)
        self.lb_frame_idx.setObjectName("lb_frame_idx")
        self.horizontalLayout_6.addWidget(self.lb_frame_idx)
        self.sb_frame_idx = QtWidgets.QSpinBox(Form)
        self.sb_frame_idx.setMinimum(0)
        self.sb_frame_idx.setMaximum(100000000)
        self.sb_frame_idx.setObjectName("sb_frame_idx")
        self.horizontalLayout_6.addWidget(self.sb_frame_idx)
        self.cb_set_frame_idx_na = QtWidgets.QCheckBox(Form)
        self.cb_set_frame_idx_na.setObjectName("cb_set_frame_idx_na")
        self.horizontalLayout_6.addWidget(self.cb_set_frame_idx_na)
        spacerItem4 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_6.addItem(spacerItem4)
        self.verticalLayout.addLayout(self.horizontalLayout_6)
        self.label_4 = QtWidgets.QLabel(Form)
        self.label_4.setObjectName("label_4")
        self.verticalLayout.addWidget(self.label_4)
        self.leComment = QtWidgets.QPlainTextEdit(Form)
        self.leComment.setObjectName("leComment")
        self.verticalLayout.addWidget(self.leComment)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem5 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem5)
        self.pbCancel = QtWidgets.QPushButton(Form)
        self.pbCancel.setObjectName("pbCancel")
        self.horizontalLayout.addWidget(self.pbCancel)
        self.pbOK = QtWidgets.QPushButton(Form)
        self.pbOK.setDefault(True)
        self.pbOK.setObjectName("pbOK")
        self.horizontalLayout.addWidget(self.pbOK)
        self.verticalLayout.addLayout(self.horizontalLayout)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Edit event"))
        self.label.setText(_translate("Form", "Time"))
        self.pb_set_to_current_time.setText(_translate("Form", "Set to current time"))
        self.cb_set_time_na.setText(_translate("Form", "Set NA"))
        self.lb_image_idx.setText(_translate("Form", "Image index"))
        self.pb_set_to_current_image_index.setText(_translate("Form", "Set to current image index"))
        self.lbSubject.setText(_translate("Form", "Subject"))
        self.label_2.setText(_translate("Form", "Code"))
        self.lb_frame_idx.setText(_translate("Form", "Frame index"))
        self.cb_set_frame_idx_na.setText(_translate("Form", "Set NA"))
        self.label_4.setText(_translate("Form", "Comment"))
        self.pbCancel.setText(_translate("Form", "Cancel"))
        self.pbOK.setText(_translate("Form", "OK"))



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2024 Olivier Friard


  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  MA 02110-1301, USA.

"""

import logging
import pathlib as pl
import shutil
from math import log2

from PyQt5.QtWidgets import QFileDialog

from . import config as cfg
from . import dialog


def deinterlace(self):
    """
    change the deinterlace status of player
    """

    logging.info("change deinterlace status of player")

    for dw in self.dw_player:
        dw.player.deinterlace = self.action_deinterlace.isChecked()


def snapshot(self):
    """
    MEDIA obs: take snapshot of current video at current position
    IMAGES obs: save a copy of the current image

    snapshot is saved on media path following the template: MEDIA-FILE-NAME_TIME-POSITION.png
    """

    if self.playerType == cfg.MEDIA:
        for i, player in enumerate(self.dw_player):
            if (
                str(i + 1) in self.pj[cfg.OBSERVATIONS][self.observationId][cfg.FILE]
                and self.pj[cfg.OBSERVATIONS][self.observationId][cfg.FILE][str(i + 1)]
            ):
                p = pl.Path(self.dw_player[0].player.playlist[self.dw_player[0].player.playlist_pos]["filename"])

                snapshot_file_path = str(p.parent / f"{p.stem}_{player.player.time_pos:0.3f}.png")

                player.player.screenshot_to_file(snapshot_file_path)
                self.statusbar.showMessage(f"Video snapshot saved in {snapshot_file_path}", 0)

                logging.debug(f"video snapshot saved in {snapshot_file_path}")

    if self.playerType == cfg.IMAGES:
        output_file_name, _ = QFileDialog().getSaveFileName(
            self, "Save copy of the current image", pl.Path(self.images_list[self.image_idx]).name
        )
        if output_file_name:
            shutil.copyfile(self.images_list[self.image_idx], output_file_name)
            self.statusbar.showMessage(f"Image saved in {output_file_name}", 0)

            logging.debug(f"video snapshot saved in {output_file_name}")


def zoom_level(self):
    """
    display dialog box for setting the zoom level
    """
    logging.info("change zoom level of player")

    players_list: list = []
    for idx, dw in enumerate(self.dw_player):
        players_list.append(("dsb", f"Player #{idx + 1}", 0.1, 12, 0.1, 2**dw.player.video_zoom, 1))

        """
        zoom_levels: list = []
        for choice in (2, 1, 0.5, 0.25):
            zoom_levels.append((str(choice), "selected" if log2(choice) == dw.player.video_zoom else ""))
        players_list.append(("il", f"Player #{idx + 1}", zoom_levels))
        """

    zl = dialog.Input_dialog(label_caption="Select the zoom level", elements_list=players_list, title="Video zoom level")
    if not zl.exec_():
        return

    if cfg.ZOOM_LEVEL not in self.pj[cfg.OBSERVATIONS][self.observationId][cfg.MEDIA_INFO]:
        self.pj[cfg.OBSERVATIONS][self.observationId][cfg.MEDIA_INFO][cfg.ZOOM_LEVEL] = {}

    for idx, dw in enumerate(self.dw_player):
        if (
            self.pj[cfg.OBSERVATIONS][self.observationId][cfg.MEDIA_INFO][cfg.ZOOM_LEVEL].get(str(idx + 1), dw.player.video_zoom)
            != zl.elements[f"Player #{idx + 1}"].value()
        ):
            dw.player.video_zoom = log2(float(zl.elements[f"Player #{idx + 1}"].value()))

            logging.debug(f"video zoom changed in {dw.player.video_zoom} for player {idx + 1}")

            self.pj[cfg.OBSERVATIONS][self.observationId][cfg.MEDIA_INFO][cfg.ZOOM_LEVEL][str(idx + 1)] = float(
                zl.elements[f"Player #{idx + 1}"].value()
            )
            display_zoom_level(self)
            self.project_changed()


def rotate_displayed_video(self):
    """
    rotate the displayed video
    """
    players_list: list = []
    for idx, dw in enumerate(self.dw_player):
        rotation_angles: list = []
        for choice in (0, 90, 180, 270):
            rotation_angles.append((str(choice), "selected" if choice == dw.player.video_rotate else ""))
        players_list.append(("il", f"Player #{idx + 1}", rotation_angles))

    w = dialog.Input_dialog(label_caption="Select the rotation angle", elements_list=players_list, title="Video rotation angle")
    if not w.exec_():
        return
    if cfg.ROTATION_ANGLE not in self.pj[cfg.OBSERVATIONS][self.observationId][cfg.MEDIA_INFO]:
        self.pj[cfg.OBSERVATIONS][self.observationId][cfg.MEDIA_INFO][cfg.ROTATION_ANGLE] = {}

    for idx, dw in enumerate(self.dw_player):
        if self.pj[cfg.OBSERVATIONS][self.observationId][cfg.MEDIA_INFO][cfg.ROTATION_ANGLE].get(
            str(idx + 1), dw.player.video_rotate
        ) != float(w.elements[f"Player #{idx + 1}"].currentText()):
            dw.player.video_rotate = int(w.elements[f"Player #{idx + 1}"].currentText())

            logging.debug(f"video rotation changed to {dw.player.video_rotate} for player {idx + 1}")

            self.pj[cfg.OBSERVATIONS][self.observationId][cfg.MEDIA_INFO][cfg.ROTATION_ANGLE][str(idx + 1)] = int(
                w.elements[f"Player #{idx + 1}"].currentText()
            )
            self.project_changed()


def display_subtitles(self):
    """
    display dialog for subtitles display
    """
    players_list = []
    for idx, dw in enumerate(self.dw_player):
        if cfg.DISPLAY_MEDIA_SUBTITLES in self.pj[cfg.OBSERVATIONS][self.observationId][cfg.MEDIA_INFO]:
            default = self.pj[cfg.OBSERVATIONS][self.observationId][cfg.MEDIA_INFO][cfg.DISPLAY_MEDIA_SUBTITLES].get(
                str(idx + 1), dw.player.sub_visibility
            )
        else:
            default = dw.player.sub_visibility
        players_list.append(("cb", f"Player #{idx + 1}", default))

    st = dialog.Input_dialog("Display subtitles", players_list)
    if not st.exec_():
        return

    if cfg.DISPLAY_MEDIA_SUBTITLES not in self.pj[cfg.OBSERVATIONS][self.observationId][cfg.MEDIA_INFO]:
        self.pj[cfg.OBSERVATIONS][self.observationId][cfg.MEDIA_INFO][cfg.DISPLAY_MEDIA_SUBTITLES] = {}

    for idx, dw in enumerate(self.dw_player):
        if (
            self.pj[cfg.OBSERVATIONS][self.observationId][cfg.MEDIA_INFO][cfg.DISPLAY_MEDIA_SUBTITLES].get(
                str(idx + 1), dw.player.sub_visibility
            )
            != st.elements[f"Player #{idx + 1}"].isChecked()
        ):
            dw.player.sub_visibility = st.elements[f"Player #{idx + 1}"].isChecked()

            logging.debug(f"subtitle visibility for player {idx + 1}: {dw.player.sub_visibility}")

            self.pj[cfg.OBSERVATIONS][self.observationId][cfg.MEDIA_INFO][cfg.DISPLAY_MEDIA_SUBTITLES][str(idx + 1)] = st.elements[
                f"Player #{idx + 1}"
            ].isChecked()
            self.project_changed()


def display_zoom_level(self) -> None:
    """
    display the zoom level
    """
    msg: str = "Zoom level: <b>"
    for player in self.dw_player:
        msg += f"{2**player.player.video_zoom:.1f} "
    msg += "</b>"
    self.lb_zoom_level.setText(msg)


def display_play_rate(self) -> None:
    """
    display current play rate in status bar widget
    """

    self.lb_video_info.setText(f"Play rate: <b>x{self.play_rate:.3f}</b>")

    logging.debug(f"play rate: {self.play_rate:.3f}")


def video_normalspeed_activated(self):
    """
    set playing speed at normal speed (1x)
    """

    if self.playerType != cfg.MEDIA:
        return

    self.play_rate = 1
    for i, player in enumerate(self.dw_player):
        if (
            str(i + 1) in self.pj[cfg.OBSERVATIONS][self.observationId][cfg.FILE]
            and self.pj[cfg.OBSERVATIONS][self.observationId][cfg.FILE][str(i + 1)]
        ):
            player.player.speed = self.play_rate

    display_play_rate(self)


def video_faster_activated(self):
    """
    increase playing speed by play_rate_step value
    """

    if self.playerType != cfg.MEDIA:
        return

    if self.play_rate + self.play_rate_step <= 60:
        self.play_rate += self.play_rate_step
        for i, player in enumerate(self.dw_player):
            if (
                str(i + 1) in self.pj[cfg.OBSERVATIONS][self.observationId][cfg.FILE]
                and self.pj[cfg.OBSERVATIONS][self.observationId][cfg.FILE][str(i + 1)]
            ):
                player.player.speed = self.play_rate

        display_play_rate(self)


def video_slower_activated(self):
    """
    decrease playing speed by play_rate_step value
    """

    if self.playerType != cfg.MEDIA:
        return

    if self.play_rate - self.play_rate_step >= 0.1:
        self.play_rate -= self.play_rate_step

        for i, player in enumerate(self.dw_player):
            player.player.speed = round(self.play_rate, 3)

        display_play_rate(self)
